# future
from __future__ import annotations

# stdlib
from collections import defaultdict
from enum import Enum
from typing import Any

# third party
from result import Err
from result import Ok
from result import Result
from typing_extensions import Self

# relative
from ..serde.serializable import serializable
from ..server.credentials import SyftVerifyKey
from ..service.action.action_permissions import ActionObjectEXECUTE
from ..service.action.action_permissions import ActionObjectOWNER
from ..service.action.action_permissions import ActionObjectPermission
from ..service.action.action_permissions import ActionObjectREAD
from ..service.action.action_permissions import ActionObjectWRITE
from ..service.action.action_permissions import ActionPermission
from ..service.action.action_permissions import StoragePermission
from ..service.context import AuthedServiceContext
from ..service.response import SyftSuccess
from ..types.syft_object import SyftObject
from ..types.uid import UID
from .document_store import BaseStash
from .document_store import PartitionKey
from .document_store import PartitionKeys
from .document_store import QueryKey
from .document_store import QueryKeys
from .document_store import StorePartition


@serializable(canonical_name="UniqueKeyCheck", version=1)
class UniqueKeyCheck(Enum):
    EMPTY = 0
    MATCHES = 1
    ERROR = 2


class KeyValueBackingStore:
    """Key-Value store core logic."""

    def __setitem__(self, key: Any, value: Any) -> None:
        raise NotImplementedError

    def __getitem__(self, key: Any) -> Self:
        raise NotImplementedError

    def __repr__(self) -> str:
        raise NotImplementedError

    def __len__(self) -> int:
        raise NotImplementedError

    def __delitem__(self, key: str) -> None:
        raise NotImplementedError

    def clear(self) -> None:
        raise NotImplementedError

    def copy(self) -> Self:
        raise NotImplementedError

    def update(self, *args: Any, **kwargs: Any) -> None:
        raise NotImplementedError

    def keys(self) -> Any:
        raise NotImplementedError

    def values(self) -> Any:
        raise NotImplementedError

    def items(self) -> Any:
        raise NotImplementedError

    def pop(self, *args: Any) -> Self:
        raise NotImplementedError

    def __contains__(self, item: Any) -> bool:
        raise NotImplementedError

    def __iter__(self) -> Any:
        raise NotImplementedError


class KeyValueStorePartition(StorePartition):
    """Key-Value StorePartition

    Parameters:
        `settings`: PartitionSettings
            PySyft specific settings
        `store_config`: StoreConfig
            Backend specific configuration
    """

    def init_store(self) -> Result[Ok, Err]:
        store_status = super().init_store()
        if store_status.is_err():
            return store_status

        try:
            self.data = self.store_config.backing_store(
                "data", self.settings, self.store_config
            )
            self.unique_keys = self.store_config.backing_store(
                "unique_keys", self.settings, self.store_config
            )
            self.searchable_keys = self.store_config.backing_store(
                "searchable_keys", self.settings, self.store_config
            )
            # uid -> set['<uid>_permission']
            self.permissions: dict[UID, set[str]] = self.store_config.backing_store(
                "permissions", self.settings, self.store_config, ddtype=set
            )

            # uid -> set['<server_uid>']
            self.storage_permissions: dict[UID, set[UID]] = (
                self.store_config.backing_store(
                    "storage_permissions",
                    self.settings,
                    self.store_config,
                    ddtype=set,
                )
            )

            for partition_key in self.unique_cks:
                pk_key = partition_key.key
                if pk_key not in self.unique_keys:
                    self.unique_keys[pk_key] = {}

            for partition_key in self.searchable_cks:
                pk_key = partition_key.key
                if pk_key not in self.searchable_keys:
                    self.searchable_keys[pk_key] = defaultdict(list)
        except BaseException as e:
            return Err(str(e))

        return Ok(True)

    def __len__(self) -> int:
        return len(self.data)

    def _get(
        self,
        uid: UID,
        credentials: SyftVerifyKey,
        has_permission: bool | None = False,
    ) -> Result[SyftObject, str]:
        # relative
        from ..service.action.action_store import ActionObjectREAD

        # if you get something you need READ permission
        read_permission = ActionObjectREAD(uid=uid, credentials=credentials)

        if self.has_permission(read_permission) or has_permission:
            syft_object = self.data[uid]
            return Ok(syft_object)
        return Err(f"Permission: {read_permission} denied")

    # Potentially thread-unsafe methods.
    # CAUTION:
    #       * Don't use self.lock here.
    #       * Do not call the public thread-safe methods here(with locking).
    # These methods are called from the public thread-safe API, and will hang the process.

    def _set(
        self,
        credentials: SyftVerifyKey,
        obj: SyftObject,
        add_permissions: list[ActionObjectPermission] | None = None,
        add_storage_permission: bool = True,
        ignore_duplicates: bool = False,
    ) -> Result[SyftObject, str]:
        try:
            # if obj.id is None:
            # obj.id = UID()
            store_query_key: QueryKey = self.settings.store_key.with_obj(obj)
            uid = store_query_key.value
            write_permission = ActionObjectWRITE(uid=uid, credentials=credentials)
            can_write = self.has_permission(write_permission)
            unique_query_keys: QueryKeys = self.settings.unique_keys.with_obj(obj)
            store_key_exists = store_query_key.value in self.data
            searchable_query_keys = self.settings.searchable_keys.with_obj(obj)

            ck_check = self._check_partition_keys_unique(
                unique_query_keys=unique_query_keys
            )

            if not store_key_exists and ck_check == UniqueKeyCheck.EMPTY:
                # attempt to claim it for writing
                ownership_result = self.take_ownership(uid=uid, credentials=credentials)
                can_write = ownership_result.is_ok()
            elif not ignore_duplicates:
                keys = ", ".join(f"`{key.key}`" for key in unique_query_keys.all)
                return Err(
                    f"Duplication Key Error for {obj}.\n"
                    f"The fields that should be unique are {keys}."
                )
            else:
                # we are not throwing an error, because we are ignoring duplicates
                # we are also not writing though
                return Ok(obj)

            if can_write:
                self._set_data_and_keys(
                    store_query_key=store_query_key,
                    unique_query_keys=unique_query_keys,
                    searchable_query_keys=searchable_query_keys,
                    obj=obj,
                )
                self.data[uid] = obj

                # Add default permissions
                if uid not in self.permissions:
                    self.permissions[uid] = set()
                self.add_permission(ActionObjectREAD(uid=uid, credentials=credentials))
                if add_permissions is not None:
                    self.add_permissions(add_permissions)

                if uid not in self.storage_permissions:
                    self.storage_permissions[uid] = set()
                if add_storage_permission:
                    self.add_storage_permission(
                        StoragePermission(
                            uid=uid,
                            server_uid=self.server_uid,
                        )
                    )

                return Ok(obj)
            else:
                return Err(f"Permission: {write_permission} denied")
        except Exception as e:
            return Err(f"Failed to write obj {obj}. {e}")

    def take_ownership(
        self, uid: UID, credentials: SyftVerifyKey
    ) -> Result[SyftSuccess, str]:
        # first person using this UID can claim ownership
        if uid not in self.permissions and uid not in self.data:
            self.add_permissions(
                [
                    ActionObjectOWNER(uid=uid, credentials=credentials),
                    ActionObjectWRITE(uid=uid, credentials=credentials),
                    ActionObjectREAD(uid=uid, credentials=credentials),
                    ActionObjectEXECUTE(uid=uid, credentials=credentials),
                ]
            )
            return Ok(SyftSuccess(message=f"Ownership of ID: {uid} taken."))
        return Err(f"UID: {uid} already owned.")

    def add_permission(self, permission: ActionObjectPermission) -> None:
        permissions = self.permissions[permission.uid]
        permissions.add(permission.permission_string)
        self.permissions[permission.uid] = permissions

    def remove_permission(self, permission: ActionObjectPermission) -> None:
        permissions = self.permissions[permission.uid]
        permissions.remove(permission.permission_string)
        self.permissions[permission.uid] = permissions

    def add_permissions(self, permissions: list[ActionObjectPermission]) -> None:
        for permission in permissions:
            self.add_permission(permission)

    def has_permission(self, permission: ActionObjectPermission) -> bool:
        if not isinstance(permission.permission, ActionPermission):
            raise Exception(f"ObjectPermission type: {permission.permission} not valid")

        if (
            permission.credentials
            and self.root_verify_key.verify == permission.credentials.verify
        ):
            return True

        if (
            permission.credentials
            and self.has_admin_permissions is not None
            and self.has_admin_permissions(permission.credentials)
        ):
            return True

        if (
            permission.uid in self.permissions
            and permission.permission_string in self.permissions[permission.uid]
        ):
            return True

        # 🟡 TODO 14: add ALL_READ, ALL_EXECUTE etc
        # third party
        if permission.permission == ActionPermission.OWNER:
            pass
        elif (
            permission.permission == ActionPermission.READ
            and ActionObjectPermission(
                permission.uid, ActionPermission.ALL_READ
            ).permission_string
            in self.permissions[permission.uid]
        ):
            return True
        elif permission.permission == ActionPermission.WRITE:
            pass
        elif permission.permission == ActionPermission.EXECUTE:
            pass

        return False

    def _get_permissions_for_uid(self, uid: UID) -> Result[set[str], Err]:
        if uid in self.permissions:
            return Ok(self.permissions[uid])
        return Err(f"No permissions found for uid: {uid}")

    def get_all_permissions(self) -> Result[dict[UID, set[str]], str]:
        return Ok(dict(self.permissions.items()))

    def add_storage_permission(self, permission: StoragePermission) -> None:
        permissions = self.storage_permissions[permission.uid]
        permissions.add(permission.server_uid)
        self.storage_permissions[permission.uid] = permissions

    def add_storage_permissions(self, permissions: list[StoragePermission]) -> None:
        for permission in permissions:
            self.add_storage_permission(permission)

    def remove_storage_permission(self, permission: StoragePermission) -> None:
        permissions = self.storage_permissions[permission.uid]
        permissions.remove(permission.server_uid)
        self.storage_permissions[permission.uid] = permissions

    def has_storage_permission(self, permission: StoragePermission | UID) -> bool:
        if isinstance(permission, UID):
            permission = StoragePermission(uid=permission, server_uid=self.server_uid)

        if permission.uid in self.storage_permissions:
            return permission.server_uid in self.storage_permissions[permission.uid]
        return False

    def _get_storage_permissions_for_uid(self, uid: UID) -> Result[set[UID], Err]:
        if uid in self.storage_permissions:
            return Ok(self.storage_permissions[uid])
        return Err(f"No storage permissions found for uid: {uid}")

    def get_all_storage_permissions(self) -> Result[dict[UID, set[UID]], str]:
        return Ok(dict(self.storage_permissions.items()))

    def _all(
        self,
        credentials: SyftVerifyKey,
        order_by: PartitionKey | None = None,
        has_permission: bool | None = False,
    ) -> Result[list[BaseStash.object_type], str]:
        # this checks permissions
        res = [self._get(uid, credentials, has_permission) for uid in self.data.keys()]
        result = [x.ok() for x in res if x.is_ok()]
        if order_by is not None:
            result = sorted(result, key=lambda x: getattr(x, order_by.key, ""))
        return Ok(result)

    def _remove_keys(
        self,
        store_key: QueryKey,
        unique_query_keys: QueryKeys,
        searchable_query_keys: QueryKeys,
    ) -> None:
        uqks = unique_query_keys.all
        for qk in uqks:
            pk_key, pk_value = qk.key, qk.value
            ck_col = self.unique_keys[pk_key]
            ck_col.pop(store_key.value, None)
            self.unique_keys[pk_key] = ck_col

        sqks = searchable_query_keys.all
        for qk in sqks:
            pk_key, pk_value = qk.key, qk.value
            ck_col = self.searchable_keys[pk_key]
            if isinstance(pk_value, list):
                for pk_value_item in pk_value:
                    pk_value_str = str(pk_value_item)
                    if pk_value_str in ck_col and (
                        store_key.value in ck_col[pk_value_str]
                    ):
                        ck_col[pk_value_str].remove(store_key.value)
            else:
                if pk_value in ck_col and (store_key.value in ck_col[pk_value]):
                    ck_col[pk_value].remove(store_key.value)
            self.searchable_keys[pk_key] = ck_col

    def _find_index_or_search_keys(
        self,
        credentials: SyftVerifyKey,
        index_qks: QueryKeys,
        search_qks: QueryKeys,
        order_by: PartitionKey | None = None,
    ) -> Result[list[SyftObject], str]:
        ids: set | None = None
        errors = []
        # third party
        if len(index_qks.all) > 0:
            index_results = self._get_keys_index(qks=index_qks)
            if index_results.is_ok():
                if ids is None:
                    ids = index_results.ok()
                ids = ids.intersection(index_results.ok())
            else:
                errors.append(index_results.err())

        search_results = None
        if len(search_qks.all) > 0:
            search_results = self._find_keys_search(qks=search_qks)

            if search_results.is_ok():
                if ids is None:
                    ids = search_results.ok()
                ids = ids.intersection(search_results.ok())
            else:
                errors.append(search_results.err())

        if len(errors) > 0:
            return Err(" ".join(errors))

        if ids is None:
            return Ok([])

        qks: QueryKeys = self.store_query_keys(ids)
        return self._get_all_from_store(
            credentials=credentials, qks=qks, order_by=order_by
        )

    def _update(
        self,
        credentials: SyftVerifyKey,
        qk: QueryKey,
        obj: SyftObject,
        has_permission: bool = False,
        overwrite: bool = False,
        allow_missing_keys: bool = False,
    ) -> Result[SyftObject, str]:
        try:
            if qk.value not in self.data:
                return Err(f"No object exists for query key: {qk}")

            if has_permission or self.has_permission(
                ActionObjectWRITE(uid=qk.value, credentials=credentials)
            ):
                _original_obj = self.data[qk.value]
                _original_unique_keys = self.settings.unique_keys.with_obj(
                    _original_obj
                )
                if allow_missing_keys:
                    searchable_keys = PartitionKeys(
                        pks=[
                            x
                            for x in self.settings.searchable_keys.all
                            if hasattr(_original_obj, x.key)
                        ]
                    )
                    _original_searchable_keys = searchable_keys.with_obj(_original_obj)

                else:
                    _original_searchable_keys = self.settings.searchable_keys.with_obj(
                        _original_obj
                    )

                store_query_key = self.settings.store_key.with_obj(_original_obj)

                # remove old keys
                self._remove_keys(
                    store_key=store_query_key,
                    unique_query_keys=_original_unique_keys,
                    searchable_query_keys=_original_searchable_keys,
                )

                # update the object with new data
                if overwrite:
                    # Overwrite existing object and their values
                    _original_obj = obj
                else:
                    for key, value in obj.to_dict(exclude_empty=True).items():
                        if key == "id":
                            # protected field
                            continue
                        setattr(_original_obj, key, value)

                # update data and keys
                self._set_data_and_keys(
                    store_query_key=store_query_key,
                    unique_query_keys=self.settings.unique_keys.with_obj(_original_obj),
                    searchable_query_keys=self.settings.searchable_keys.with_obj(
                        _original_obj
                    ),
                    # has been updated
                    obj=_original_obj,
                )

                # 🟡 TODO 28: Add locking in this transaction

                return Ok(_original_obj)
            else:
                return Err(f"Failed to update obj {obj}, you have no permission")

        except Exception as e:
            return Err(f"Failed to update obj {obj} with error: {e}")

    def _get_all_from_store(
        self,
        credentials: SyftVerifyKey,
        qks: QueryKeys,
        order_by: PartitionKey | None = None,
    ) -> Result[list[SyftObject], str]:
        matches = []
        for qk in qks.all:
            if qk.value in self.data:
                if self.has_permission(
                    ActionObjectREAD(uid=qk.value, credentials=credentials)
                ):
                    matches.append(self.data[qk.value])
        if order_by is not None:
            matches = sorted(matches, key=lambda x: getattr(x, order_by.key, ""))
        return Ok(matches)

    def create(self, obj: SyftObject) -> Result[SyftObject, str]:
        pass

    def _delete(
        self, credentials: SyftVerifyKey, qk: QueryKey, has_permission: bool = False
    ) -> Result[SyftSuccess, Err]:
        try:
            if has_permission or self.has_permission(
                ActionObjectWRITE(uid=qk.value, credentials=credentials)
            ):
                _obj = self.data.pop(qk.value)
                self.permissions.pop(qk.value)
                self.storage_permissions.pop(qk.value)
                self._delete_unique_keys_for(_obj)
                self._delete_search_keys_for(_obj)
                return Ok(SyftSuccess(message="Deleted"))
            else:
                return Err(
                    f"Failed to delete with query key {qk}, you have no permission"
                )
        except Exception as e:
            return Err(f"Failed to delete with query key {qk} with error: {e}")

    def _delete_unique_keys_for(self, obj: SyftObject) -> Result[SyftSuccess, str]:
        for _unique_ck in self.unique_cks:
            qk = _unique_ck.with_obj(obj)
            unique_keys = self.unique_keys[qk.key]
            unique_keys.pop(qk.value, None)
            self.unique_keys[qk.key] = unique_keys
        return Ok(SyftSuccess(message="Deleted"))

    def _delete_search_keys_for(self, obj: SyftObject) -> Result[SyftSuccess, str]:
        for _search_ck in self.searchable_cks:
            qk: QueryKey = _search_ck.with_obj(obj)
            search_keys: defaultdict = self.searchable_keys[qk.key]
            if isinstance(qk.value, list):
                for qk_value in qk.value:
                    search_keys.pop(qk_value, None)
            else:
                search_keys.pop(qk.value, None)
            self.searchable_keys[qk.key] = search_keys
        return Ok(SyftSuccess(message="Deleted"))

    def _get_keys_index(self, qks: QueryKeys) -> Result[set[Any], str]:
        try:
            # match AND
            subsets: list = []
            for qk in qks.all:
                subset: set = set()
                pk_key, pk_value = qk.key, qk.value
                if pk_key not in self.unique_keys:
                    return Err(f"Failed to query index with {qk}")
                ck_col = self.unique_keys[pk_key]
                if pk_value not in ck_col.keys():
                    # must be at least one in all query keys
                    continue
                store_value = ck_col[pk_value]
                subsets.append({store_value})

            if len(subsets) == 0:
                return Ok(set())
            # AND
            subset = subsets.pop()
            for s in subsets:
                subset = subset.intersection(s)

            return Ok(subset)
        except Exception as e:
            return Err(f"Failed to query with {qks}. {e}")

    def _find_keys_search(self, qks: QueryKeys) -> Result[set[QueryKey], str]:
        try:
            # match AND
            subsets = []
            for qk in qks.all:
                subset: set = set()
                pk_key, pk_value = qk.key, qk.value
                if pk_key not in self.searchable_keys:
                    return Err(f"Failed to search with {qk}")
                ck_col = self.searchable_keys[pk_key]
                if qk.type_list:
                    # 🟡 TODO: change this hacky way to do on to many relationships
                    # this is when you search a QueryKey which is a list of items
                    # at the moment its mostly just a List[UID]
                    # match OR against all keys for this col
                    # the values of the list will be turned into strings in a single key
                    matches = set()
                    for item in pk_value:
                        for col_key in ck_col.keys():
                            if str(item) in col_key:
                                store_values = ck_col[col_key]
                                for value in store_values:
                                    matches.add(value)
                    if len(matches):
                        subsets.append(matches)
                else:
                    # this is the normal path
                    if pk_value not in ck_col.keys():
                        # must be at least one in all query keys
                        subsets.append(set())
                        continue
                    store_values = ck_col[pk_value]
                    subsets.append(set(store_values))

            if len(subsets) == 0:
                return Ok(set())
            # AND
            subset = subsets.pop()
            for s in subsets:
                subset = subset.intersection(s)
            return Ok(subset)
        except Exception as e:
            return Err(f"Failed to query with {qks}. {e}")

    def _check_partition_keys_unique(
        self, unique_query_keys: QueryKeys
    ) -> UniqueKeyCheck:
        # dont check the store key
        qks = [
            x
            for x in unique_query_keys.all
            if x.partition_key != self.settings.store_key
        ]
        matches = []
        for qk in qks:
            pk_key, pk_value = qk.key, qk.value
            if pk_key not in self.unique_keys:
                raise Exception(
                    f"pk_key: {pk_key} not in unique_keys: {self.unique_keys.keys()}"
                )
            ck_col = self.unique_keys[pk_key]
            if pk_value in ck_col:
                matches.append(pk_key)

        if len(matches) == 0:
            return UniqueKeyCheck.EMPTY
        elif len(matches) == len(qks):
            return UniqueKeyCheck.MATCHES

        return UniqueKeyCheck.ERROR

    def _set_data_and_keys(
        self,
        store_query_key: QueryKey,
        unique_query_keys: QueryKeys,
        searchable_query_keys: QueryKeys,
        obj: SyftObject,
    ) -> None:
        uqks = unique_query_keys.all

        for qk in uqks:
            pk_key, pk_value = qk.key, qk.value
            ck_col = self.unique_keys[pk_key]
            ck_col[pk_value] = store_query_key.value
            self.unique_keys[pk_key] = ck_col

        self.unique_keys[store_query_key.key][store_query_key.value] = (
            store_query_key.value
        )

        sqks = searchable_query_keys.all
        for qk in sqks:
            pk_key, pk_value = qk.key, qk.value
            ck_col = self.searchable_keys[pk_key]
            if qk.type_list:
                # coerce the list of objects to strings for a single key
                pk_value = " ".join([str(obj) for obj in pk_value])

            # check if key is present, then add to existing key
            if pk_value in ck_col:
                ck_col[pk_value].append(store_query_key.value)
            else:
                # else create the key with a list
                ck_col[pk_value] = [store_query_key.value]

            self.searchable_keys[pk_key] = ck_col

        self.data[store_query_key.value] = obj

    def _migrate_data(
        self, to_klass: SyftObject, context: AuthedServiceContext, has_permission: bool
    ) -> Result[bool, str]:
        credentials = context.credentials
        has_permission = (credentials == self.root_verify_key) or has_permission
        if has_permission:
            for key, value in self.data.items():
                try:
                    migrated_value = value.migrate_to(to_klass.__version__, context)
                except Exception:
                    return Err(
                        f"Failed to migrate data to {to_klass} for qk {to_klass.__version__}: {key}"
                    )
                qk = self.settings.store_key.with_obj(key)
                result = self._update(
                    credentials,
                    qk=qk,
                    obj=migrated_value,
                    has_permission=has_permission,
                    overwrite=True,
                    allow_missing_keys=True,
                )

                if result.is_err():
                    return result.err()

            return Ok(True)

        return Err("You don't have permissions to migrate data.")



# stdlib
from collections import defaultdict
import sys
from typing import cast

# third party
from result import Err
from result import Ok
from result import Result

# relative
from ...serde.serializable import serializable
from ...store.document_store import DocumentStore
from ...store.document_store import StorePartition
from ...types.blob_storage import BlobStorageEntry
from ...types.syft_object import SyftObject
from ..action.action_object import Action
from ..action.action_object import ActionObject
from ..action.action_permissions import ActionObjectPermission
from ..action.action_permissions import StoragePermission
from ..action.action_store import KeyValueActionStore
from ..context import AuthedServiceContext
from ..response import SyftError
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import service_method
from ..user.user_roles import ADMIN_ROLE_LEVEL
from .object_migration_state import MigrationData
from .object_migration_state import StoreMetadata
from .object_migration_state import SyftMigrationStateStash
from .object_migration_state import SyftObjectMigrationState


@serializable(canonical_name="MigrationService", version=1)
class MigrationService(AbstractService):
    store: DocumentStore
    stash: SyftMigrationStateStash

    def __init__(self, store: DocumentStore) -> None:
        self.store = store
        self.stash = SyftMigrationStateStash(store=store)

    @service_method(path="migration", name="get_version")
    def get_version(
        self, context: AuthedServiceContext, canonical_name: str
    ) -> int | SyftError:
        """Search for the metadata for an object."""

        result = self.stash.get_by_name(
            canonical_name=canonical_name, credentials=context.credentials
        )

        if result.is_err():
            return SyftError(message=f"{result.err()}")

        migration_state = result.ok()

        if migration_state is None:
            return SyftError(
                message=f"No migration state exists for canonical name: {canonical_name}"
            )

        return migration_state.current_version

    @service_method(path="migration", name="get_state")
    def get_state(
        self, context: AuthedServiceContext, canonical_name: str
    ) -> bool | SyftError:
        result = self.stash.get_by_name(
            canonical_name=canonical_name, credentials=context.credentials
        )

        if result.is_err():
            return SyftError(message=f"{result.err()}")

        return result.ok()

    @service_method(path="migration", name="register_migration_state")
    def register_migration_state(
        self,
        context: AuthedServiceContext,
        current_version: int,
        canonical_name: str,
    ) -> SyftObjectMigrationState | SyftError:
        obj = SyftObjectMigrationState(
            current_version=current_version, canonical_name=canonical_name
        )
        result = self.stash.set(migration_state=obj, credentials=context.credentials)

        if result.is_err():
            return SyftError(message=f"{result.err()}")

        return result.ok()

    def _find_klasses_pending_for_migration(
        self, context: AuthedServiceContext, object_types: list[type[SyftObject]]
    ) -> list[type[SyftObject]]:
        klasses_to_be_migrated = []

        for object_type in object_types:
            canonical_name = object_type.__canonical_name__
            object_version = object_type.__version__

            migration_state = self.get_state(context, canonical_name)
            if isinstance(migration_state, SyftError):
                raise Exception(
                    f"Failed to get migration state for {canonical_name}. Error: {migration_state}"
                )
            if (
                migration_state is not None
                and migration_state.current_version != migration_state.latest_version
            ):
                klasses_to_be_migrated.append(object_type)
            else:
                self.register_migration_state(
                    context,
                    current_version=object_version,
                    canonical_name=canonical_name,
                )

        return klasses_to_be_migrated

    @service_method(
        path="migration.get_all_store_metadata",
        name="get_all_store_metadata",
        roles=ADMIN_ROLE_LEVEL,
    )
    def get_all_store_metadata(
        self,
        context: AuthedServiceContext,
        document_store_object_types: list[type[SyftObject]] | None = None,
        include_action_store: bool = True,
    ) -> dict[str, StoreMetadata] | SyftError:
        res = self._get_all_store_metadata(
            context,
            document_store_object_types=document_store_object_types,
            include_action_store=include_action_store,
        )
        if res.is_err():
            return SyftError(message=res.value)
        else:
            return res.ok()

    def _get_partition_from_type(
        self,
        context: AuthedServiceContext,
        object_type: type[SyftObject],
    ) -> Result[KeyValueActionStore | StorePartition, str]:
        object_partition: KeyValueActionStore | StorePartition | None = None
        if issubclass(object_type, ActionObject):
            object_partition = cast(KeyValueActionStore, context.server.action_store)
        else:
            canonical_name = object_type.__canonical_name__  # type: ignore[unreachable]
            object_partition = self.store.partitions.get(canonical_name)

        if object_partition is None:
            return Err(f"Object partition not found for {object_type}")  # type: ignore

        return Ok(object_partition)

    def _get_store_metadata(
        self,
        context: AuthedServiceContext,
        object_type: type[SyftObject],
    ) -> Result[StoreMetadata, str]:
        object_partition = self._get_partition_from_type(context, object_type)
        if object_partition.is_err():
            return object_partition
        object_partition = object_partition.ok()

        permissions = object_partition.get_all_permissions()

        if permissions.is_err():
            return permissions
        permissions = permissions.ok()

        storage_permissions = object_partition.get_all_storage_permissions()
        if storage_permissions.is_err():
            return storage_permissions
        storage_permissions = storage_permissions.ok()

        return Ok(
            StoreMetadata(
                object_type=object_type,
                permissions=permissions,
                storage_permissions=storage_permissions,
            )
        )

    def _get_all_store_metadata(
        self,
        context: AuthedServiceContext,
        document_store_object_types: list[type[SyftObject]] | None = None,
        include_action_store: bool = True,
    ) -> Result[dict[str, list[str]], str]:
        if document_store_object_types is None:
            document_store_object_types = self.store.get_partition_object_types()

        store_metadata = {}
        for klass in document_store_object_types:
            result = self._get_store_metadata(context, klass)
            if result.is_err():
                return result
            store_metadata[klass] = result.ok()

        if include_action_store:
            result = self._get_store_metadata(context, ActionObject)
            if result.is_err():
                return result
            store_metadata[ActionObject] = result.ok()

        return Ok(store_metadata)

    @service_method(
        path="migration.update_store_metadata",
        name="update_store_metadata",
        roles=ADMIN_ROLE_LEVEL,
    )
    def update_store_metadata(
        self, context: AuthedServiceContext, store_metadata: dict[type, StoreMetadata]
    ) -> SyftSuccess | SyftError:
        res = self._update_store_metadata(context, store_metadata)
        if res.is_err():
            return SyftError(message=res.value)
        else:
            return SyftSuccess(message=res.ok())

    def _update_store_metadata_for_klass(
        self, context: AuthedServiceContext, metadata: StoreMetadata
    ) -> Result[str, str]:
        object_partition = self._get_partition_from_type(context, metadata.object_type)
        if object_partition.is_err():
            return object_partition
        object_partition = object_partition.ok()

        permissions = [
            ActionObjectPermission.from_permission_string(uid, perm_str)
            for uid, perm_strs in metadata.permissions.items()
            for perm_str in perm_strs
        ]

        storage_permissions = [
            StoragePermission(uid, server_uid)
            for uid, server_uids in metadata.storage_permissions.items()
            for server_uid in server_uids
        ]

        object_partition.add_permissions(permissions)
        object_partition.add_storage_permissions(storage_permissions)

        return Ok("success")

    def _update_store_metadata(
        self, context: AuthedServiceContext, store_metadata: dict[type, StoreMetadata]
    ) -> Result[str, str]:
        print("Updating store metadata")
        for metadata in store_metadata.values():
            result = self._update_store_metadata_for_klass(context, metadata)
            if result.is_err():
                return result
        return Ok("success")

    @service_method(
        path="migration.get_migration_objects",
        name="get_migration_objects",
        roles=ADMIN_ROLE_LEVEL,
    )
    def get_migration_objects(
        self,
        context: AuthedServiceContext,
        document_store_object_types: list[type[SyftObject]] | None = None,
        get_all: bool = False,
    ) -> dict | SyftError:
        res = self._get_migration_objects(context, document_store_object_types, get_all)
        if res.is_err():
            return SyftError(message=res.value)
        else:
            return res.ok()

    def _get_migration_objects(
        self,
        context: AuthedServiceContext,
        document_store_object_types: list[type[SyftObject]] | None = None,
        get_all: bool = False,
    ) -> Result[dict[type[SyftObject], list[SyftObject]], str]:
        if document_store_object_types is None:
            document_store_object_types = self.store.get_partition_object_types()

        if get_all:
            klasses_to_migrate = document_store_object_types
        else:
            klasses_to_migrate = self._find_klasses_pending_for_migration(
                context=context, object_types=document_store_object_types
            )

        result = defaultdict(list)

        for klass in klasses_to_migrate:
            canonical_name = klass.__canonical_name__
            object_partition = self.store.partitions.get(canonical_name)
            if object_partition is None:
                continue
            objects_result = object_partition.all(
                context.credentials, has_permission=True
            )
            if objects_result.is_err():
                return objects_result
            objects = objects_result.ok()
            for object in objects:
                actual_klass = type(object)
                use_klass = (
                    klass
                    if actual_klass.__canonical_name__ == klass.__canonical_name__
                    else actual_klass
                )
                result[use_klass].append(object)

        return Ok(dict(result))

    def _search_partition_for_object(
        self, context: AuthedServiceContext, obj: SyftObject
    ) -> Result[StorePartition, str]:
        klass = type(obj)
        mro = klass.__mro__
        class_index = 0
        object_partition = None
        while len(mro) > class_index:
            canonical_name = mro[class_index].__canonical_name__
            object_partition = self.store.partitions.get(canonical_name)
            if object_partition is not None:
                break
            class_index += 1
        if object_partition is None:
            return Err(f"Object partition not found for {klass}")
        return Ok(object_partition)

    @service_method(
        path="migration.create_migrated_objects",
        name="create_migrated_objects",
        roles=ADMIN_ROLE_LEVEL,
    )
    def create_migrated_objects(
        self,
        context: AuthedServiceContext,
        migrated_objects: list[SyftObject],
        ignore_existing: bool = True,
    ) -> SyftSuccess | SyftError:
        res = self._create_migrated_objects(context, migrated_objects)
        if res.is_err():
            return SyftError(message=res.value)
        else:
            return SyftSuccess(message=res.ok())

    def _create_migrated_objects(
        self,
        context: AuthedServiceContext,
        migrated_objects: list[SyftObject],
        ignore_existing: bool = True,
    ) -> Result[str, str]:
        for migrated_object in migrated_objects:
            object_partition_or_err = self._search_partition_for_object(
                context, migrated_object
            )
            if object_partition_or_err.is_err():
                return object_partition_or_err
            object_partition = object_partition_or_err.ok()

            # upsert the object
            result = object_partition.set(
                context.credentials,
                obj=migrated_object,
            )
            if result.is_err():
                if ignore_existing and "Duplication Key Error" in result.value:
                    print(
                        f"{type(migrated_object)} #{migrated_object.id} already exists"
                    )
                    continue
                else:
                    return result

        return Ok(value="success")

    @service_method(
        path="migration.update_migrated_objects",
        name="update_migrated_objects",
        roles=ADMIN_ROLE_LEVEL,
    )
    def update_migrated_objects(
        self, context: AuthedServiceContext, migrated_objects: list[SyftObject]
    ) -> SyftSuccess | SyftError:
        res = self._update_migrated_objects(context, migrated_objects)
        if res.is_err():
            return SyftError(message=res.value)
        else:
            return SyftSuccess(message=res.ok())

    def _update_migrated_objects(
        self, context: AuthedServiceContext, migrated_objects: list[SyftObject]
    ) -> Result[str, str]:
        for migrated_object in migrated_objects:
            object_partition_or_err = self._search_partition_for_object(
                context, migrated_object
            )
            if object_partition_or_err.is_err():
                return object_partition_or_err
            object_partition = object_partition_or_err.ok()

            # canonical_name = mro[class_index].__canonical_name__
            # object_partition = self.store.partitions.get(canonical_name)

            # print(klass, canonical_name, object_partition)
            qk = object_partition.settings.store_key.with_obj(migrated_object.id)
            result = object_partition._update(
                context.credentials,
                qk=qk,
                obj=migrated_object,
                has_permission=True,
                overwrite=True,
                allow_missing_keys=True,
            )

            if result.is_err():
                print("ERR:", result.value, file=sys.stderr)
                print("ERR:", type(migrated_object), file=sys.stderr)
                print("ERR:", migrated_object, file=sys.stderr)
                # return result
        return Ok(value="success")

    def _migrate_objects(
        self,
        context: AuthedServiceContext,
        migration_objects: dict[type[SyftObject], list[SyftObject]],
    ) -> Result[list[SyftObject], str]:
        migrated_objects = []
        for klass, objects in migration_objects.items():
            canonical_name = klass.__canonical_name__
            # Migrate data for objects in document store
            print(f"Migrating data for: {canonical_name} table.")
            for object in objects:
                try:
                    migrated_value = object.migrate_to(klass.__version__, context)
                    migrated_objects.append(migrated_value)
                except Exception:
                    # stdlib
                    import traceback

                    print(traceback.format_exc())
                    return Err(
                        f"Failed to migrate data to {klass} for qk {klass.__version__}: {object.id}"
                    )
        return Ok(migrated_objects)

    @service_method(
        path="migration.migrate_data",
        name="migrate_data",
        roles=ADMIN_ROLE_LEVEL,
    )
    def migrate_data(
        self,
        context: AuthedServiceContext,
        document_store_object_types: list[type[SyftObject]] | None = None,
    ) -> SyftSuccess | SyftError:
        # Track all object type that need migration for document store

        # get all objects, keyed by type (because we might want to have different rules for different types)
        # Q: will this be tricky with the protocol????
        # A: For now we will assume that the client will have the same version

        # Then, locally we write stuff that says
        # for klass, objects in migration_dict.items():
        # for object in objects:
        #   if isinstance(object, X):
        #        do something custom
        #   else:
        #       migrated_value = object.migrate_to(klass.__version__, context)
        #
        # migrated_values = [SyftObject]
        # client.migration.write_migrated_values(migrated_values)

        migration_objects_result = self._get_migration_objects(
            context, document_store_object_types
        )
        if migration_objects_result.is_err():
            return migration_objects_result
        migration_objects = migration_objects_result.ok()

        migrated_objects_result = self._migrate_objects(context, migration_objects)
        if migrated_objects_result.is_err():
            return SyftError(message=migrated_objects_result.err())
        migrated_objects = migrated_objects_result.ok()

        objects_update_update_result = self._update_migrated_objects(
            context, migrated_objects
        )
        if objects_update_update_result.is_err():
            return SyftError(message=objects_update_update_result.value)

        migration_actionobjects_result = self._get_migration_actionobjects(context)

        if migration_actionobjects_result.is_err():
            return SyftError(message=migration_actionobjects_result.err())
        migration_actionobjects = migration_actionobjects_result.ok()

        migrated_actionobjects = self._migrate_objects(context, migration_actionobjects)
        if migrated_actionobjects.is_err():
            return SyftError(message=migrated_actionobjects.err())
        migrated_actionobjects = migrated_actionobjects.ok()

        actionobjects_update_update_result = self._update_migrated_actionobjects(
            context, migrated_actionobjects
        )
        if actionobjects_update_update_result.is_err():
            return SyftError(message=actionobjects_update_update_result.err())

        return SyftSuccess(message="Data upgraded to the latest version")

    @service_method(
        path="migration.get_migration_actionobjects",
        name="get_migration_actionobjects",
        roles=ADMIN_ROLE_LEVEL,
    )
    def get_migration_actionobjects(
        self, context: AuthedServiceContext, get_all: bool = False
    ) -> dict | SyftError:
        res = self._get_migration_actionobjects(context, get_all=get_all)
        if res.is_ok():
            return res.ok()
        else:
            return SyftError(message=res.value)

    def _get_migration_actionobjects(
        self, context: AuthedServiceContext, get_all: bool = False
    ) -> Result[dict[type[SyftObject], list[SyftObject]], str]:
        # Track all object types from action store
        action_object_types = [Action, ActionObject]
        action_object_types.extend(ActionObject.__subclasses__())
        klass_by_canonical_name: dict[str, type[SyftObject]] = {
            klass.__canonical_name__: klass for klass in action_object_types
        }

        action_object_pending_migration = self._find_klasses_pending_for_migration(
            context=context, object_types=action_object_types
        )
        result_dict: dict[type[SyftObject], list[SyftObject]] = defaultdict(list)
        action_store = context.server.action_store
        action_store_objects_result = action_store._all(
            context.credentials, has_permission=True
        )
        if action_store_objects_result.is_err():
            return action_store_objects_result
        action_store_objects = action_store_objects_result.ok()

        for obj in action_store_objects:
            if get_all or type(obj) in action_object_pending_migration:
                klass = klass_by_canonical_name.get(obj.__canonical_name__, type(obj))
                result_dict[klass].append(obj)  # type: ignore
        return Ok(dict(result_dict))

    @service_method(
        path="migration.update_migrated_actionobjects",
        name="update_migrated_actionobjects",
        roles=ADMIN_ROLE_LEVEL,
    )
    def update_migrated_actionobjects(
        self, context: AuthedServiceContext, objects: list[SyftObject]
    ) -> SyftSuccess | SyftError:
        res = self._update_migrated_actionobjects(context, objects)
        if res.is_ok():
            return SyftSuccess(message="succesfully migrated actionobjects")
        else:
            return SyftError(message=res.value)

    def _update_migrated_actionobjects(
        self, context: AuthedServiceContext, objects: list[SyftObject]
    ) -> Result[str, str]:
        # Track all object types from action store
        action_store = context.server.action_store
        for obj in objects:
            res = action_store.set(
                uid=obj.id, credentials=context.credentials, syft_object=obj
            )
            if res.is_err():
                return res
        return Ok("success")

    @service_method(
        path="migration.get_migration_data",
        name="get_migration_data",
        roles=ADMIN_ROLE_LEVEL,
    )
    def get_migration_data(
        self, context: AuthedServiceContext
    ) -> MigrationData | SyftError:
        store_objects_result = self._get_migration_objects(context, get_all=True)
        if store_objects_result.is_err():
            return SyftError(message=store_objects_result.err())
        store_objects = store_objects_result.ok()

        action_objects_result = self._get_migration_actionobjects(context, get_all=True)
        if action_objects_result.is_err():
            return SyftError(message=action_objects_result.err())
        action_objects = action_objects_result.ok()

        blob_storage_objects = store_objects.pop(BlobStorageEntry, [])

        store_metadata_result = self._get_all_store_metadata(context)
        if store_metadata_result.is_err():
            return SyftError(message=store_metadata_result.err())
        store_metadata = store_metadata_result.ok()

        return MigrationData(
            server_uid=context.server.id,
            signing_key=context.server.signing_key,
            store_objects=store_objects,
            metadata=store_metadata,
            action_objects=action_objects,
            blob_storage_objects=blob_storage_objects,
        )

    @service_method(
        path="migration.apply_migration_data",
        name="apply_migration_data",
        roles=ADMIN_ROLE_LEVEL,
    )
    def apply_migration_data(
        self,
        context: AuthedServiceContext,
        migration_data: MigrationData,
    ) -> SyftSuccess | SyftError:
        # NOTE blob storage is migrated via client,
        # it needs access to both source and destination blob storages.
        if len(migration_data.blobs):
            return SyftError(
                message="Blob storage migration is not supported by this endpoint, "
                "please use 'client.load_migration_data' instead."
            )

        # migrate + apply store objects
        migrated_objects_result = self._migrate_objects(
            context, migration_data.store_objects
        )
        if migrated_objects_result.is_err():
            return SyftError(message=migrated_objects_result.err())
        migrated_objects = migrated_objects_result.ok()
        store_objects_result = self._create_migrated_objects(context, migrated_objects)
        if store_objects_result.is_err():
            return SyftError(message=store_objects_result.err())

        # migrate+apply action objects
        migrated_actionobjects = self._migrate_objects(
            context, migration_data.action_objects
        )
        if migrated_actionobjects.is_err():
            return SyftError(message=migrated_actionobjects.err())
        migrated_actionobjects = migrated_actionobjects.ok()
        action_objects_result = self._update_migrated_actionobjects(
            context, migrated_actionobjects
        )
        if action_objects_result.is_err():
            return SyftError(message=action_objects_result.err())

        # apply metadata
        metadata_result = self._update_store_metadata(context, migration_data.metadata)
        if metadata_result.is_err():
            return SyftError(message=metadata_result.err())

        return SyftSuccess(message="Migration completed successfully")



# stdlib
from io import BytesIO
from pathlib import Path
import sys
from typing import Any

# third party
from result import Result
from typing_extensions import Self
import yaml

# relative
from ...serde.deserialize import _deserialize
from ...serde.serializable import serializable
from ...serde.serialize import _serialize
from ...server.credentials import SyftSigningKey
from ...server.credentials import SyftVerifyKey
from ...store.document_store import BaseStash
from ...store.document_store import DocumentStore
from ...store.document_store import PartitionKey
from ...store.document_store import PartitionSettings
from ...types.blob_storage import BlobStorageEntry
from ...types.blob_storage import CreateBlobStorageEntry
from ...types.syft_object import Context
from ...types.syft_object import SYFT_OBJECT_VERSION_1
from ...types.syft_object import SyftBaseObject
from ...types.syft_object import SyftObject
from ...types.syft_object_registry import SyftObjectRegistry
from ...types.uid import UID
from ...util.util import prompt_warning_message
from ..action.action_permissions import ActionObjectPermission
from ..response import SyftError
from ..response import SyftSuccess


@serializable()
class SyftObjectMigrationState(SyftObject):
    __canonical_name__ = "SyftObjectMigrationState"
    __version__ = SYFT_OBJECT_VERSION_1

    __attr_unique__ = ["canonical_name"]

    canonical_name: str
    current_version: int

    @property
    def latest_version(self) -> int | None:
        available_versions = SyftObjectRegistry.get_versions(
            canonical_name=self.canonical_name,
        )
        if not available_versions:
            return None

        return sorted(available_versions, reverse=True)[0]

    @property
    def supported_versions(self) -> list:
        return SyftObjectRegistry.get_versions(self.canonical_name)


KlassNamePartitionKey = PartitionKey(key="canonical_name", type_=str)


@serializable(canonical_name="SyftMigrationStateStash", version=1)
class SyftMigrationStateStash(BaseStash):
    object_type = SyftObjectMigrationState
    settings: PartitionSettings = PartitionSettings(
        name=SyftObjectMigrationState.__canonical_name__,
        object_type=SyftObjectMigrationState,
    )

    def __init__(self, store: DocumentStore) -> None:
        super().__init__(store=store)

    def set(
        self,
        credentials: SyftVerifyKey,
        migration_state: SyftObjectMigrationState,
        add_permissions: list[ActionObjectPermission] | None = None,
        add_storage_permission: bool = True,
        ignore_duplicates: bool = False,
    ) -> Result[SyftObjectMigrationState, str]:
        res = self.check_type(migration_state, self.object_type)
        # we dont use and_then logic here as it is hard because of the order of the arguments
        if res.is_err():
            return res
        return super().set(
            credentials=credentials,
            obj=res.ok(),
            add_permissions=add_permissions,
            add_storage_permission=add_storage_permission,
            ignore_duplicates=ignore_duplicates,
        )

    def get_by_name(
        self, canonical_name: str, credentials: SyftVerifyKey
    ) -> Result[SyftObjectMigrationState, str]:
        qks = KlassNamePartitionKey.with_obj(canonical_name)
        return self.query_one(credentials=credentials, qks=qks)


@serializable()
class StoreMetadata(SyftBaseObject):
    __canonical_name__ = "StoreMetadata"
    __version__ = SYFT_OBJECT_VERSION_1

    object_type: type
    permissions: dict[UID, set[str]]
    storage_permissions: dict[UID, set[UID]]


@serializable()
class MigrationData(SyftObject):
    __canonical_name__ = "MigrationData"
    __version__ = SYFT_OBJECT_VERSION_1

    server_uid: UID
    signing_key: SyftSigningKey
    store_objects: dict[type[SyftObject], list[SyftObject]]
    metadata: dict[type[SyftObject], StoreMetadata]
    action_objects: dict[type[SyftObject], list[SyftObject]]
    blob_storage_objects: list[SyftObject]
    blobs: dict[UID, Any] = {}

    __repr_attrs__ = [
        "server_uid",
        "root_verify_key",
        "num_objects",
        "num_action_objects",
        "includes_blobs",
    ]

    @property
    def root_verify_key(self) -> SyftVerifyKey:
        return self.signing_key.verify_key

    @property
    def num_objects(self) -> int:
        return sum(len(objs) for objs in self.store_objects.values())

    @property
    def num_action_objects(self) -> int:
        return sum(len(objs) for objs in self.action_objects.values())

    @property
    def includes_blobs(self) -> bool:
        blob_ids = [obj.id for obj in self.blob_storage_objects]
        return set(self.blobs.keys()) == set(blob_ids)

    def make_migration_config(self) -> dict[str, Any]:
        server_uid = self.server_uid.to_string()
        server_private_key = str(self.signing_key)
        migration_config = {
            "server": {
                "env": [
                    {"name": "SERVER_UID", "value": server_uid},
                    {"name": "SERVER_PRIVATE_KEY", "value": server_private_key},
                ]
            }
        }
        return migration_config

    @classmethod
    def from_file(self, path: str | Path) -> Self | SyftError:
        path = Path(path)
        if not path.exists():
            return SyftError(f"File {str(path)} does not exist.")

        with open(path, "rb") as f:
            res: MigrationData = _deserialize(f.read(), from_bytes=True)

        return res

    def save(self, path: str | Path, yaml_path: str | Path) -> SyftSuccess | SyftError:
        if not self.includes_blobs:
            proceed = prompt_warning_message(
                "You are saving migration data without blob storage data. "
                "This means that any existing blobs will be missing when you load this data."
                "\nTo include blobs, call `download_blobs()` before saving.",
                confirm=True,
            )
            if not proceed:
                return SyftError(message="Migration data not saved.")

        path = Path(path)
        with open(path, "wb") as f:
            f.write(_serialize(self, to_bytes=True))

        yaml_path = Path(yaml_path)
        migration_config = self.make_migration_config()
        with open(yaml_path, "w") as f:
            yaml.dump(migration_config, f)

        return SyftSuccess(message=f"Migration data saved to {str(path)}.")

    def download_blobs(self) -> None | SyftError:
        for obj in self.blob_storage_objects:
            blob = self.download_blob(obj.id)
            if isinstance(blob, SyftError):
                return blob
            self.blobs[obj.id] = blob
        return None

    def download_blob(self, obj_id: str) -> Any | SyftError:
        api = self._get_api()
        if isinstance(api, SyftError):
            return api

        blob_retrieval = api.services.blob_storage.read(obj_id)
        if isinstance(blob_retrieval, SyftError):
            return blob_retrieval
        return blob_retrieval.read()

    def migrate_and_upload_blobs(self) -> SyftSuccess | SyftError:
        for obj in self.blob_storage_objects:
            upload_result = self.migrate_and_upload_blob(obj)
            if isinstance(upload_result, SyftError):
                return upload_result
        return SyftSuccess(message="All blobs uploaded successfully.")

    def migrate_and_upload_blob(self, obj: BlobStorageEntry) -> SyftSuccess | SyftError:
        api = self._get_api()
        if isinstance(api, SyftError):
            return api

        if obj.id not in self.blobs:
            return SyftError(f"Blob {obj.id} not found in migration data.")
        data = self.blobs[obj.id]

        migrated_obj = obj.migrate_to(BlobStorageEntry.__version__, Context())
        serialized = _serialize(data, to_bytes=True)
        size = sys.getsizeof(serialized)
        blob_create = CreateBlobStorageEntry.from_blob_storage_entry(migrated_obj)
        blob_create.file_size = size
        blob_deposit_object = api.services.blob_storage.allocate_for_user(
            blob_create, migrated_obj.uploaded_by
        )

        if isinstance(blob_deposit_object, SyftError):
            return blob_deposit_object
        return blob_deposit_object.write(BytesIO(serialized))

    def copy_without_blobs(self) -> "MigrationData":
        # Create a shallow copy of the MigrationData instance, removing blob-related data
        # This is required for sending the MigrationData to the backend.
        copy_data = self.__class__(
            server_uid=self.server_uid,
            signing_key=self.signing_key,
            store_objects=self.store_objects.copy(),
            metadata=self.metadata.copy(),
            action_objects=self.action_objects.copy(),
            blob_storage_objects=[],
            blobs={},
        )
        return copy_data



# stdlib
import contextlib
import logging
import os
from pathlib import Path
import socket
import socketserver
import sys
from typing import Any

# third party
import docker
from docker.models.containers import Container
from kr8s.objects import Pod

# relative
from ...abstract_server import AbstractServer
from ...custom_worker.builder import CustomWorkerBuilder
from ...custom_worker.builder_types import ImageBuildResult
from ...custom_worker.builder_types import ImagePushResult
from ...custom_worker.config import PrebuiltWorkerConfig
from ...custom_worker.k8s import KubeUtils
from ...custom_worker.k8s import PodStatus
from ...custom_worker.runner_k8s import KubernetesRunner
from ...server.credentials import SyftVerifyKey
from ...types.uid import UID
from ...util.util import get_queue_address
from ..response import SyftError
from .image_identifier import SyftWorkerImageIdentifier
from .worker_image import SyftWorkerImage
from .worker_image_stash import SyftWorkerImageStash
from .worker_pool import ContainerSpawnStatus
from .worker_pool import SyftWorker
from .worker_pool import WorkerHealth
from .worker_pool import WorkerOrchestrationType
from .worker_pool import WorkerStatus

logger = logging.getLogger(__name__)

DEFAULT_WORKER_IMAGE_TAG = "openmined/default-worker-image-cpu:0.0.1"
DEFAULT_WORKER_POOL_NAME = "default-pool"
K8S_SERVER_CREDS_NAME = "server-creds"


def backend_container_name() -> str:
    hostname = socket.gethostname()
    service_name = os.getenv("SERVICE", "backend")
    return f"{hostname}-{service_name}-1"


def get_container(
    docker_client: docker.DockerClient, container_name: str
) -> Container | None:
    try:
        existing_container = docker_client.containers.get(container_name)
    except docker.errors.NotFound:
        existing_container = None

    return existing_container


def extract_config_from_backend(
    worker_name: str, docker_client: docker.DockerClient
) -> dict[str, Any]:
    # Existing main backend container
    backend_container = get_container(
        docker_client, container_name=backend_container_name()
    )

    # Config with defaults
    extracted_config: dict[str, Any] = {
        "volume_binds": {},
        "network_mode": None,
        "environment": {},
    }

    if backend_container is None:
        return extracted_config

    # Inspect the existing container
    details = backend_container.attrs

    host_config = details["HostConfig"]
    environment = details["Config"]["Env"]

    # Extract Volume Binds
    vol_binds = {}

    # ignore any irrelevant binds for the worker like
    # packages/grid/backend/grid:/root/app/grid
    # packages/syft:/root/app/syft
    # packages/grid/data/package-cache:/root/.cache
    valid_binds = {
        "/var/run/docker.sock",
        "/root/.cache",
    }

    for vol in host_config["Binds"]:
        parts = vol.split(":")
        key = parts[0]
        bind = parts[1]
        mode = parts[2]

        if "/root/data/creds" in vol:
            # we need this because otherwise we are using the same server private key
            # which will make account creation fail
            key = f"{key}-{worker_name}"
        elif bind not in valid_binds:
            continue

        vol_binds[key] = {"bind": bind, "mode": mode}

    # Extract Environment Variables
    extracted_config["environment"] = dict([e.split("=", 1) for e in environment])
    extracted_config["network_mode"] = f"container:{backend_container.id}"
    extracted_config["volume_binds"] = vol_binds

    return extracted_config


def get_free_tcp_port() -> int:
    with socketserver.TCPServer(("localhost", 0), None) as s:
        free_port = s.server_address[1]
        return free_port


def run_container_using_docker(
    worker_image: SyftWorkerImage,
    docker_client: docker.DockerClient,
    worker_name: str,
    worker_count: int,
    pool_name: str,
    queue_port: int,
    debug: bool = False,
    username: str | None = None,
    password: str | None = None,
    registry_url: str | None = None,
) -> ContainerSpawnStatus:
    if not worker_image.is_built:
        raise ValueError("Image must be built before running it.")

    # Get hostname
    hostname = socket.gethostname()

    # Create a random UID for Worker
    syft_worker_uid = UID()

    # container name
    container_name = f"{hostname}-{worker_name}"

    # start container
    container = None
    error_message = None
    worker = None
    try:
        # login to the registry through the client
        # so that the subsequent pull/run commands work
        if registry_url and username and password:
            docker_client.login(
                username=username,
                password=password,
                registry=registry_url,
            )

        # If container with given name already exists then stop it
        # and recreate it.
        existing_container = get_container(
            docker_client=docker_client,
            container_name=container_name,
        )
        if existing_container:
            existing_container.remove(force=True)

        # Extract Config from backend container
        backend_host_config = extract_config_from_backend(
            worker_name=worker_name, docker_client=docker_client
        )

        # Create List of Envs to pass
        environment = backend_host_config["environment"]
        environment["CREATE_PRODUCER"] = "false"
        environment["N_CONSUMERS"] = 1
        environment["PORT"] = str(get_free_tcp_port())
        environment["CONSUMER_SERVICE_NAME"] = pool_name
        environment["SYFT_WORKER_UID"] = syft_worker_uid
        environment["DEV_MODE"] = debug
        environment["QUEUE_PORT"] = queue_port
        environment["CONTAINER_HOST"] = "docker"

        if worker_image.image_identifier is None:
            raise ValueError(f"Image {worker_image} does not have an identifier")

        container = docker_client.containers.run(
            image=worker_image.image_identifier.full_name_with_tag,
            name=f"{hostname}-{worker_name}",
            detach=True,
            # auto_remove=True,
            network_mode=backend_host_config["network_mode"],
            environment=environment,
            volumes=backend_host_config["volume_binds"],
            tty=True,
            stdin_open=True,
            labels={"orgs.openmined.syft": "this is a syft worker container"},
        )

        status = (
            WorkerStatus.STOPPED
            if container.status == "exited"
            else WorkerStatus.PENDING
        )

        healthcheck: WorkerHealth = _get_healthcheck_based_on_status(status)

        worker = SyftWorker(
            id=syft_worker_uid,
            name=worker_name,
            container_id=container.id,
            status=status,
            healthcheck=healthcheck,
            image=worker_image,
            worker_pool_name=pool_name,
        )
    except Exception as e:
        error_message = f"Failed to run command in container. {worker_name} {worker_image}. {e}. {sys.stderr}"
        if container:
            worker = SyftWorker(
                name=worker_name,
                container_id=container.id,
                status=WorkerStatus.STOPPED,
                healthcheck=WorkerHealth.UNHEALTHY,
                image=worker_image,
                worker_pool_name=pool_name,
            )
            container.stop()

    return ContainerSpawnStatus(
        worker_name=worker_name, worker=worker, error=error_message
    )


def run_workers_in_threads(
    server: AbstractServer,
    pool_name: str,
    number: int,
    start_idx: int = 0,
) -> list[ContainerSpawnStatus]:
    results = []

    for worker_count in range(start_idx + 1, number + 1):
        error = None
        worker_name = f"{pool_name}-{worker_count}"
        worker = SyftWorker(
            name=worker_name,
            status=WorkerStatus.RUNNING,
            worker_pool_name=pool_name,
            healthcheck=WorkerHealth.HEALTHY,
        )
        try:
            port = server.queue_config.client_config.queue_port
            address = get_queue_address(port)
            server.add_consumer_for_service(
                service_name=pool_name,
                syft_worker_id=worker.id,
                address=address,
            )
        except Exception as e:
            logger.error(
                f"Failed to start consumer for pool={pool_name} worker={worker_name}",
                exc_info=e,
            )
            worker.status = WorkerStatus.STOPPED
            error = str(e)

        container_status = ContainerSpawnStatus(
            worker_name=worker_name,
            worker=worker,
            error=error,
        )

        results.append(container_status)

    return results


def prepare_kubernetes_pool_env(
    runner: KubernetesRunner, env_vars: dict
) -> tuple[list, dict]:
    # get current backend pod name
    backend_pod_name = os.getenv("K8S_POD_NAME")
    if not backend_pod_name:
        raise ValueError("Pod name not provided in environment variable")

    # get current backend's credentials path
    creds_path: str | Path | None = os.getenv("CREDENTIALS_PATH")
    if not creds_path:
        raise ValueError("Credentials path not provided")

    creds_path = Path(creds_path)
    if creds_path is not None and not creds_path.exists():
        raise ValueError("Credentials file does not exist")

    # create a secret for the server credentials owned by the backend, not the pool.
    server_secret = KubeUtils.create_secret(
        secret_name=K8S_SERVER_CREDS_NAME,
        type="Opaque",
        component=backend_pod_name,
        data={creds_path.name: creds_path.read_text()},
        encoded=False,
    )

    # clone and patch backend environment variables
    backend_env = runner.get_pod_env_vars(backend_pod_name) or []
    env_vars_: list = KubeUtils.patch_env_vars(backend_env, env_vars)
    mount_secrets = {
        server_secret.metadata.name: {
            "mountPath": str(creds_path),
            "subPath": creds_path.name,
        },
    }

    return env_vars_, mount_secrets


def create_kubernetes_pool(
    runner: KubernetesRunner,
    tag: str,
    pool_name: str,
    replicas: int,
    queue_port: int,
    debug: bool,
    registry_username: str | None = None,
    registry_password: str | None = None,
    reg_url: str | None = None,
    pod_annotations: dict[str, str] | None = None,
    pod_labels: dict[str, str] | None = None,
    **kwargs: Any,
) -> list[Pod] | SyftError:
    pool = None

    try:
        logger.info(f"Creating new pool name={pool_name} tag={tag} replicas={replicas}")

        env_vars, mount_secrets = prepare_kubernetes_pool_env(
            runner,
            {
                "SYFT_WORKER": "True",
                "DEV_MODE": f"{debug}",
                "QUEUE_PORT": f"{queue_port}",
                "CONSUMER_SERVICE_NAME": pool_name,
                "N_CONSUMERS": "1",
                "CREATE_PRODUCER": "False",
                "INMEMORY_WORKERS": "False",
            },
        )

        # run the pool with args + secret
        pool = runner.create_pool(
            pool_name=pool_name,
            tag=tag,
            replicas=replicas,
            env_vars=env_vars,
            mount_secrets=mount_secrets,
            registry_username=registry_username,
            registry_password=registry_password,
            reg_url=reg_url,
            pod_annotations=pod_annotations,
            pod_labels=pod_labels,
        )
    except Exception as e:
        if pool:
            try:
                pool.delete()  # this raises another exception if the pool never starts
            except Exception as e2:
                logger.error(
                    f"Failed to delete pool {pool_name} after failed creation. {e2}"
                )
        # stdlib
        import traceback

        return SyftError(
            message=f"Failed to start workers {e} {e.__class__} {e.args} {traceback.format_exc()}."
        )

    return runner.get_pool_pods(pool_name=pool_name)


def scale_kubernetes_pool(
    runner: KubernetesRunner,
    pool_name: str,
    replicas: int,
) -> list[Pod] | SyftError:
    pool = runner.get_pool(pool_name)
    if not pool:
        return SyftError(message=f"Pool does not exist. name={pool_name}")

    try:
        logger.info(f"Scaling pool name={pool_name} to replicas={replicas}")
        runner.scale_pool(pool_name=pool_name, replicas=replicas)
    except Exception as e:
        return SyftError(message=f"Failed to scale workers {e}")

    return runner.get_pool_pods(pool_name=pool_name)


def run_workers_in_kubernetes(
    worker_image: SyftWorkerImage,
    worker_count: int,
    pool_name: str,
    queue_port: int,
    start_idx: int = 0,
    debug: bool = False,
    registry_username: str | None = None,
    registry_password: str | None = None,
    reg_url: str | None = None,
    pod_annotations: dict[str, str] | None = None,
    pod_labels: dict[str, str] | None = None,
    **kwargs: Any,
) -> list[ContainerSpawnStatus] | SyftError:
    spawn_status = []
    runner = KubernetesRunner()

    if not runner.exists(pool_name=pool_name):
        if worker_image.image_identifier is not None:
            pool_pods = create_kubernetes_pool(
                runner=runner,
                tag=worker_image.image_identifier.full_name_with_tag,
                pool_name=pool_name,
                replicas=worker_count,
                queue_port=queue_port,
                debug=debug,
                registry_username=registry_username,
                registry_password=registry_password,
                reg_url=reg_url,
                pod_annotations=pod_annotations,
                pod_labels=pod_labels,
            )
        else:
            return SyftError(
                message=f"image with uid {worker_image.id} does not have an image identifier"
            )
    else:
        pool_pods = scale_kubernetes_pool(runner, pool_name, worker_count)

        if isinstance(pool_pods, list) and len(pool_pods) > 0:
            # slice only those pods that we're interested in
            pool_pods = pool_pods[start_idx:]

    if isinstance(pool_pods, SyftError):
        return pool_pods

    # create worker object
    for pod in pool_pods:
        status: PodStatus | WorkerStatus | None = runner.get_pod_status(pod)
        status, healthcheck, error = map_pod_to_worker_status(status)

        # this worker id will be the same as the one in the worker
        syft_worker_uid = UID.with_seed(pod.metadata.name)

        worker = SyftWorker(
            id=syft_worker_uid,
            name=pod.metadata.name,
            container_id=None,
            status=status,
            healthcheck=healthcheck,
            image=worker_image,
            worker_pool_name=pool_name,
        )

        spawn_status.append(
            ContainerSpawnStatus(
                worker_name=pod.metadata.name,
                worker=worker,
                error=error,
            )
        )

    return spawn_status


def map_pod_to_worker_status(
    status: PodStatus,
) -> tuple[WorkerStatus, WorkerHealth, str | None]:
    worker_status = None
    worker_healthcheck = None
    worker_error = None

    # check if pod is ready through pod.status.condition.Ready & pod.status.condition.ContainersReady
    pod_ready = status.condition.ready and status.condition.containers_ready

    if not pod_ready:
        # extract error if not ready
        worker_error = f"{status.container.reason}: {status.container.message}"

    # map readiness to status - it's either running or pending.
    # closely relates to pod.status.phase, but avoiding as it is not as detailed as pod.status.conditions
    worker_status = WorkerStatus.RUNNING if pod_ready else WorkerStatus.PENDING

    # TODO: update these values based on actual runtime probes instead of kube pod statuses
    # if there are any errors, then healthcheck is unhealthy
    worker_healthcheck = (
        WorkerHealth.UNHEALTHY if worker_error else WorkerHealth.HEALTHY
    )

    return worker_status, worker_healthcheck, worker_error


def run_containers(
    pool_name: str,
    worker_image: SyftWorkerImage,
    number: int,
    orchestration: WorkerOrchestrationType,
    queue_port: int,
    dev_mode: bool = False,
    start_idx: int = 0,
    registry_username: str | None = None,
    registry_password: str | None = None,
    reg_url: str | None = None,
    pod_annotations: dict[str, str] | None = None,
    pod_labels: dict[str, str] | None = None,
) -> list[ContainerSpawnStatus] | SyftError:
    results = []

    if not worker_image.is_built:
        return SyftError(message="Image must be built before running it.")

    logger.info(f"Starting workers with start_idx={start_idx} count={number}")

    if orchestration == WorkerOrchestrationType.DOCKER:
        with contextlib.closing(docker.from_env()) as client:
            for worker_count in range(start_idx + 1, number + 1):
                worker_name = f"{pool_name}-{worker_count}"
                spawn_result = run_container_using_docker(
                    docker_client=client,
                    worker_name=worker_name,
                    worker_count=worker_count,
                    worker_image=worker_image,
                    pool_name=pool_name,
                    queue_port=queue_port,
                    debug=dev_mode,
                    username=registry_username,
                    password=registry_password,
                    registry_url=reg_url,
                )
                results.append(spawn_result)
    elif orchestration == WorkerOrchestrationType.KUBERNETES:
        return run_workers_in_kubernetes(
            worker_image=worker_image,
            worker_count=number,
            pool_name=pool_name,
            queue_port=queue_port,
            debug=dev_mode,
            start_idx=start_idx,
            registry_username=registry_username,
            registry_password=registry_password,
            reg_url=reg_url,
            pod_annotations=pod_annotations,
            pod_labels=pod_labels,
        )

    return results


def create_default_image(
    credentials: SyftVerifyKey,
    image_stash: SyftWorkerImageStash,
    tag: str,
    in_kubernetes: bool = False,
) -> SyftError | SyftWorkerImage:
    if not in_kubernetes:
        tag = f"openmined/syft-backend:{tag}"

    worker_config = PrebuiltWorkerConfig(
        tag=tag,
        description="Prebuilt default worker image",
    )

    # create SyftWorkerImage from a pre-built image
    _new_image = SyftWorkerImage(
        config=worker_config,
        created_by=credentials,
        image_identifier=SyftWorkerImageIdentifier.from_str(tag),
    )

    result = image_stash.get_by_worker_config(
        credentials=credentials,
        config=worker_config,
    )

    if result.ok() is None:
        result = image_stash.set(credentials, _new_image)
        if result.is_err():
            return SyftError(message=f"Failed to save image stash: {result.err()}")

    default_syft_image = result.ok()

    return default_syft_image


def _get_healthcheck_based_on_status(status: WorkerStatus) -> WorkerHealth:
    if status in [WorkerStatus.PENDING, WorkerStatus.RUNNING]:
        return WorkerHealth.HEALTHY
    else:
        return WorkerHealth.UNHEALTHY


def image_build(
    image: SyftWorkerImage, **kwargs: dict[str, Any]
) -> ImageBuildResult | SyftError:
    if image.image_identifier is not None:
        full_tag = image.image_identifier.full_name_with_tag
        try:
            builder = CustomWorkerBuilder()
            return builder.build_image(
                config=image.config,
                tag=full_tag,
                **kwargs,
            )
        except docker.errors.APIError as e:
            return SyftError(
                message=f"Docker API error when building '{full_tag}'. Reason - {e}"
            )
        except docker.errors.DockerException as e:
            return SyftError(
                message=f"Docker exception when building '{full_tag}'. Reason - {e}"
            )
        except Exception as e:
            return SyftError(
                message=f"Unknown exception when building '{full_tag}'. Reason - {e}"
            )
    else:
        return SyftError(
            message=f"image with uid {image.id} does not have an image identifier"
        )


def image_push(
    image: SyftWorkerImage,
    username: str | None = None,
    password: str | None = None,
) -> ImagePushResult | SyftError:
    if image.image_identifier is not None:
        full_tag = image.image_identifier.full_name_with_tag
        try:
            builder = CustomWorkerBuilder()
            result = builder.push_image(
                # this should be consistent with docker build command
                tag=image.image_identifier.full_name_with_tag,
                registry_url=image.image_identifier.registry_host,
                username=username,
                password=password,
            )

            if "error" in result.logs.lower() or result.exit_code:
                return SyftError(
                    message=f"Failed to push {full_tag}. "
                    f"Exit code: {result.exit_code}. "
                    f"Logs:\n{result.logs}"
                )

            return result
        except docker.errors.APIError as e:
            return SyftError(message=f"Docker API error when pushing {full_tag}. {e}")
        except docker.errors.DockerException as e:
            return SyftError(
                message=f"Docker exception when pushing {full_tag}. Reason - {e}"
            )
        except Exception as e:
            return SyftError(
                message=f"Unknown exception when pushing {image.image_identifier}. Reason - {e}"
            )
    else:
        return SyftError(
            message=f"image with uid {image.id} does not have an "
            "image identifier and tag, hence we can't push it."
        )


def get_orchestration_type() -> WorkerOrchestrationType:
    """Returns orchestration type from env. Defaults to Python."""

    orchstration_type_ = os.getenv("CONTAINER_HOST")
    return (
        WorkerOrchestrationType(orchstration_type_.lower())
        if orchstration_type_
        else WorkerOrchestrationType.PYTHON
    )



# stdlib
from hashlib import sha256
from pathlib import Path
from secrets import token_hex
from typing import Any

# third party
from kr8s.objects import ConfigMap
from kr8s.objects import Job
from kr8s.objects import Secret

# relative
from .builder_types import BUILD_IMAGE_TIMEOUT_SEC
from .builder_types import BuilderBase
from .builder_types import ImageBuildResult
from .builder_types import ImagePushResult
from .builder_types import PUSH_IMAGE_TIMEOUT_SEC
from .k8s import INTERNAL_REGISTRY_HOST
from .k8s import KUBERNETES_NAMESPACE
from .k8s import KubeUtils
from .k8s import USE_INTERNAL_REGISTRY
from .k8s import get_kr8s_client
from .utils import ImageUtils

__all__ = ["KubernetesBuilder"]


class BuildFailed(Exception):
    pass


class KubernetesBuilder(BuilderBase):
    # app.kubernetes.io/component
    COMPONENT = "builder"

    # service account for the Job, useful for workload identity
    SERVICE_ACCOUNT = "builder-service-account"

    # Time after which Job will be deleted
    JOB_COMPLETION_TTL = 60

    def __init__(self) -> None:
        self.client = get_kr8s_client()

    def build_image(
        self,
        tag: str,
        dockerfile: str | None = None,
        dockerfile_path: Path | None = None,
        buildargs: dict | None = None,
        **kwargs: Any,
    ) -> ImageBuildResult:
        image_digest = None
        logs = None
        config = None
        job_id = self._new_job_id(tag)
        kaniko_extra_args = None

        if dockerfile:
            pass
        elif dockerfile_path:
            dockerfile = dockerfile_path.read_text()

        if USE_INTERNAL_REGISTRY:
            # if we are using internal registry, we tweak the tag to point to it
            tag = ImageUtils.change_registry(tag, registry=INTERNAL_REGISTRY_HOST)

            # and let kaniko know about the internal registry
            kaniko_extra_args = [
                f"--insecure-registry={INTERNAL_REGISTRY_HOST}",
                f"--skip-tls-verify-registry={INTERNAL_REGISTRY_HOST}",
            ]

        try:
            # Create a ConfigMap with the Dockerfile
            config = self._create_build_config(job_id, dockerfile)
            config.refresh()

            # Create and start the job
            job = self._create_kaniko_build_job(
                job_id=job_id,
                tag=tag,
                job_config=config,
                build_args=buildargs,
                kaniko_extra_args=kaniko_extra_args,
            )

            # wait for job to complete/fail
            job.wait(
                ["condition=Complete", "condition=Failed"],
                timeout=BUILD_IMAGE_TIMEOUT_SEC,
            )

            # get logs
            logs = self._get_logs(job)

            image_digest = self._get_image_digest(job)
            if not image_digest:
                exit_code = self._get_exit_code(job)
                raise BuildFailed(
                    "Failed to build the image. "
                    f"Kaniko exit code={exit_code}. "
                    f"Logs={logs}"
                )

        except Exception:
            raise
        finally:
            # don't delete the job, kubernetes will gracefully do that for us
            config and config.delete(propagation_policy="Foreground")

        return ImageBuildResult(
            image_hash=image_digest,
            logs=logs,
        )

    def push_image(
        self,
        tag: str,
        registry_url: str,
        username: str | None = None,
        password: str | None = None,
        **kwargs: Any,
    ) -> ImagePushResult:
        exit_code = 1
        logs = None
        job_id = self._new_job_id(tag)
        push_secret = None
        registry_auths = []

        if USE_INTERNAL_REGISTRY:
            # local registry auth can be anything
            registry_auths.append((INTERNAL_REGISTRY_HOST, "admin", token_hex(4)))
        else:
            # kaniko has already pushed the image directly
            return ImagePushResult(logs="Already pushed", exit_code=0)

        # if we have external registry credentials, add them to the list
        # elsea leave it for workload identity
        if username and password:
            registry_auths.append((registry_url, username, password))

        try:
            push_secret = self._create_push_secret(
                job_id=job_id,
                registry_auths=registry_auths,
            )
            push_secret.refresh()

            job = self._create_push_job(
                job_id=job_id,
                tag=tag,
                push_secret=push_secret,
            )

            job.wait(
                ["condition=Complete", "condition=Failed"],
                timeout=PUSH_IMAGE_TIMEOUT_SEC,
            )
            exit_code = self._get_exit_code(job)[0]
            logs = self._get_logs(job)
        except Exception:
            raise
        finally:
            push_secret and push_secret.delete(propagation_policy="Foreground")

        return ImagePushResult(logs=logs, exit_code=exit_code)

    def _new_job_id(self, tag: str) -> str:
        return self._get_tag_hash(tag)[:16]

    def _get_tag_hash(self, tag: str) -> str:
        return sha256(tag.encode()).hexdigest()

    def _get_image_digest(self, job: Job) -> str | None:
        selector = {"batch.kubernetes.io/job-name": job.metadata.name}
        pods = self.client.get("pods", label_selector=selector)
        return KubeUtils.get_container_exit_message(pods)

    def _get_exit_code(self, job: Job) -> list[int]:
        selector = {"batch.kubernetes.io/job-name": job.metadata.name}
        pods = self.client.get("pods", label_selector=selector)
        return KubeUtils.get_container_exit_code(pods)

    def _get_logs(self, job: Job) -> str:
        selector = {"batch.kubernetes.io/job-name": job.metadata.name}
        pods = self.client.get("pods", label_selector=selector)
        return KubeUtils.get_logs(pods)

    def _create_build_config(self, job_id: str, dockerfile: str) -> ConfigMap:
        config_map = ConfigMap(
            {
                "metadata": {
                    "name": f"build-{job_id}",
                    "labels": {
                        "app.kubernetes.io/name": KUBERNETES_NAMESPACE,
                        "app.kubernetes.io/component": KubernetesBuilder.COMPONENT,
                        "app.kubernetes.io/managed-by": "kr8s",
                    },
                },
                "data": {
                    "Dockerfile": dockerfile,
                },
            }
        )
        return KubeUtils.create_or_get(config_map)

    def _create_kaniko_build_job(
        self,
        job_id: str,
        tag: str,
        job_config: ConfigMap,
        build_args: dict | None = None,
        kaniko_extra_args: list[str] | None = None,
    ) -> Job:
        # for push
        build_args = build_args or {}
        kaniko_extra_args = kaniko_extra_args or []
        build_args_list = []

        for k, v in build_args.items():
            build_args_list.append(f'--build-arg="{k}={v}"')

        job = Job(
            {
                "metadata": {
                    "name": f"build-{job_id}",
                    "labels": {
                        "app.kubernetes.io/name": KUBERNETES_NAMESPACE,
                        "app.kubernetes.io/component": KubernetesBuilder.COMPONENT,
                        "app.kubernetes.io/managed-by": "kr8s",
                    },
                },
                "spec": {
                    "backoffLimit": 0,
                    "ttlSecondsAfterFinished": KubernetesBuilder.JOB_COMPLETION_TTL,
                    "template": {
                        "spec": {
                            "restartPolicy": "Never",
                            "serviceAccountName": KubernetesBuilder.SERVICE_ACCOUNT,
                            "containers": [
                                {
                                    "name": "kaniko",
                                    "image": "gcr.io/kaniko-project/executor:latest",
                                    "args": [
                                        "--dockerfile=Dockerfile",
                                        "--context=dir:///workspace",
                                        f"--destination={tag}",
                                        # Disabling --reproducible because it eats up a lot of CPU+RAM
                                        # https://github.com/GoogleContainerTools/kaniko/issues/1960
                                        # https://github.com/GoogleContainerTools/kaniko/pull/2477
                                        # "--reproducible",
                                        # Cache
                                        "--cache=true",
                                        "--cache-copy-layers",
                                        "--cache-run-layers",
                                        # outputs args
                                        "--digest-file=/dev/termination-log",
                                        # other kaniko conf
                                        "--log-format=text",
                                        "--verbosity=info",
                                    ]
                                    + kaniko_extra_args
                                    + build_args_list,
                                    "volumeMounts": [
                                        {
                                            "name": "build-input",
                                            "mountPath": "/workspace",
                                        },
                                    ],
                                }
                            ],
                            "volumes": [
                                {
                                    "name": "build-input",
                                    "configMap": {
                                        "name": job_config.metadata.name,
                                    },
                                },
                            ],
                        }
                    },
                },
            }
        )

        return KubeUtils.create_or_get(job)

    def _create_push_job(
        self,
        job_id: str,
        tag: str,
        push_secret: Secret,
    ) -> Job:
        internal_tag = ImageUtils.change_registry(tag, registry=INTERNAL_REGISTRY_HOST)
        internal_reg, internal_repo, _ = ImageUtils.parse_tag(internal_tag)

        run_cmds = [
            # push with credentials
            "echo Pushing image...",
            f"krane copy {internal_tag} {tag}",
            # cleanup image from internal registry
            "echo Cleaning up...",
            f"IMG_DIGEST=$(krane digest {internal_tag})",
            f"krane delete {internal_reg}/{internal_repo}@$IMG_DIGEST; echo Done",
        ]

        job = Job(
            {
                "metadata": {
                    # there should be only one push at a time, so keep this name unique to a push
                    "name": f"push-{job_id}",
                    "labels": {
                        "app.kubernetes.io/name": KUBERNETES_NAMESPACE,
                        "app.kubernetes.io/component": KubernetesBuilder.COMPONENT,
                        "app.kubernetes.io/managed-by": "kr8s",
                    },
                },
                "spec": {
                    "backoffLimit": 0,
                    "ttlSecondsAfterFinished": KubernetesBuilder.JOB_COMPLETION_TTL,
                    "template": {
                        "spec": {
                            "restartPolicy": "Never",
                            "serviceAccountName": KubernetesBuilder.SERVICE_ACCOUNT,
                            "containers": [
                                {
                                    "name": "crane",
                                    # debug is needed for "sh" to be available
                                    "image": "gcr.io/go-containerregistry/krane:debug",
                                    "command": ["sh"],
                                    "args": ["-c", " && ".join(run_cmds)],
                                    "volumeMounts": [
                                        {
                                            "name": "push-secret",
                                            "mountPath": "/root/.docker/config.json",
                                            "subPath": "config.json",
                                            "readOnly": True,
                                        },
                                    ],
                                }
                            ],
                            "volumes": [
                                {
                                    "name": "push-secret",
                                    "secret": {
                                        "secretName": push_secret.metadata.name,
                                        "items": [
                                            {
                                                "key": ".dockerconfigjson",
                                                "path": "config.json",
                                            },
                                        ],
                                    },
                                },
                            ],
                        }
                    },
                },
            }
        )
        return KubeUtils.create_or_get(job)

    def _create_push_secret(
        self,
        job_id: str,
        registry_auths: list[tuple[str, str, str]],
    ) -> Secret:
        return KubeUtils.create_dockerconfig_secret(
            secret_name=f"push-secret-{job_id}",
            component=KubernetesBuilder.COMPONENT,
            registries=registry_auths,
        )



# stdlib
from datetime import date
from datetime import datetime
from datetime import time
import functools
from importlib.util import find_spec
from io import BytesIO

# third party
from dateutil import parser
from nacl.signing import SigningKey
from nacl.signing import VerifyKey
import numpy as np
from pandas import DataFrame
from pandas import Series
from pandas._libs.tslibs.timestamps import Timestamp
import pyarrow as pa
import pyarrow.parquet as pq
import pydantic
from pydantic._internal._model_construction import ModelMetaclass
from pymongo.collection import Collection
from result import Err
from result import Ok

# relative
from ..types.dicttuple import DictTuple
from ..types.dicttuple import _Meta as _DictTupleMetaClass
from ..types.syft_metaclass import EmptyType
from ..types.syft_metaclass import PartialModelMetaclass
from .array import numpy_deserialize
from .array import numpy_serialize
from .deserialize import _deserialize as deserialize
from .recursive_primitives import _serialize_kv_pairs
from .recursive_primitives import deserialize_kv
from .recursive_primitives import deserialize_type
from .recursive_primitives import recursive_serde_register
from .recursive_primitives import recursive_serde_register_type
from .recursive_primitives import serialize_type
from .serialize import _serialize as serialize

recursive_serde_register(
    SigningKey,
    serialize=lambda x: bytes(x),
    deserialize=lambda x: SigningKey(x),
    canonical_name="nacl_signing_key",
    version=1,
)

recursive_serde_register(
    VerifyKey,
    serialize=lambda x: bytes(x),
    deserialize=lambda x: VerifyKey(x),
    canonical_name="nacl_verify_key",
    version=1,
)


# result Ok and Err
recursive_serde_register(
    Ok, serialize_attrs=["_value"], canonical_name="result_Ok", version=1
)
recursive_serde_register(
    Err, serialize_attrs=["_value"], canonical_name="result_Err", version=1
)

# exceptions
recursive_serde_register(cls=TypeError, canonical_name="TypeError", version=1)

# mongo collection
recursive_serde_register_type(
    Collection, canonical_name="pymongo_collection", version=1
)


def serialize_dataframe(df: DataFrame) -> bytes:
    table = pa.Table.from_pandas(df)
    sink = pa.BufferOutputStream()
    # 🟡 TODO 37: Should we warn about this?
    parquet_args = {
        "coerce_timestamps": "us",
        "allow_truncated_timestamps": True,
    }
    pq.write_table(table, sink, **parquet_args)
    buffer = sink.getvalue()
    numpy_bytes = buffer.to_pybytes()
    return numpy_bytes


def deserialize_dataframe(buf: bytes) -> DataFrame:
    reader = pa.BufferReader(buf)
    numpy_bytes = reader.read_buffer()
    result = pq.read_table(numpy_bytes)
    df = result.to_pandas()
    return df


# pandas
recursive_serde_register(
    DataFrame,
    serialize=serialize_dataframe,
    deserialize=deserialize_dataframe,
    canonical_name="pandas_dataframe",
    version=1,
)


def deserialize_series(blob: bytes) -> Series:
    df: DataFrame = DataFrame.from_dict(deserialize(blob, from_bytes=True))
    return Series(df[df.columns[0]])


recursive_serde_register(
    Series,
    serialize=lambda x: serialize(DataFrame(x).to_dict(), to_bytes=True),
    deserialize=deserialize_series,
    canonical_name="pandas_series",
    version=1,
)

recursive_serde_register(
    datetime,
    serialize=lambda x: serialize(x.isoformat(), to_bytes=True),
    deserialize=lambda x: parser.isoparse(deserialize(x, from_bytes=True)),
    canonical_name="datetime_datetime",
    version=1,
)

recursive_serde_register(
    time,
    serialize=lambda x: serialize(x.isoformat(), to_bytes=True),
    deserialize=lambda x: parser.parse(deserialize(x, from_bytes=True)).time(),
    canonical_name="datetime_time",
    version=1,
)

recursive_serde_register(
    date,
    serialize=lambda x: serialize(x.isoformat(), to_bytes=True),
    deserialize=lambda x: parser.parse(deserialize(x, from_bytes=True)).date(),
    canonical_name="datetime_date",
    version=1,
)

recursive_serde_register(
    Timestamp,
    serialize=lambda x: serialize(x.value, to_bytes=True),
    deserialize=lambda x: Timestamp(deserialize(x, from_bytes=True)),
    canonical_name="pandas_timestamp",
    version=1,
)


def _serialize_dicttuple(x: DictTuple) -> bytes:
    return _serialize_kv_pairs(size=len(x), kv_pairs=zip(x.keys(), x))


recursive_serde_register(
    _DictTupleMetaClass,
    serialize=serialize_type,
    deserialize=deserialize_type,
    canonical_name="dicttuple_meta",
    version=1,
)
recursive_serde_register(
    DictTuple,
    serialize=_serialize_dicttuple,
    deserialize=functools.partial(deserialize_kv, DictTuple),
    canonical_name="dicttuple",
    version=1,
)


recursive_serde_register(
    EmptyType,
    serialize=serialize_type,
    deserialize=deserialize_type,
    canonical_name="empty_type",
    version=1,
)


recursive_serde_register_type(
    ModelMetaclass, canonical_name="pydantic_model_metaclass", version=1
)
recursive_serde_register_type(
    PartialModelMetaclass, canonical_name="partial_model_metaclass", version=1
)


def serialize_bytes_io(io: BytesIO) -> bytes:
    io.seek(0)
    return serialize(io.read(), to_bytes=True)


recursive_serde_register(
    BytesIO,
    serialize=serialize_bytes_io,
    deserialize=lambda x: BytesIO(deserialize(x, from_bytes=True)),
    canonical_name="bytes_io",
    version=1,
)

try:
    # third party
    from IPython.display import Image

    recursive_serde_register(Image, canonical_name="IPython_display_Image", version=1)

except Exception:  # nosec
    pass


try:
    # third party
    import torch
    from torch._C import _TensorMeta

    recursive_serde_register_type(
        _TensorMeta, canonical_name="torch_tensor_meta", version=1
    )
    recursive_serde_register_type(
        torch.Tensor, canonical_name="torch_tensor", version=1
    )

    def torch_serialize(tensor: torch.Tensor) -> bytes:
        return numpy_serialize(tensor.numpy())

    def torch_deserialize(buffer: bytes) -> torch.tensor:
        np_array = numpy_deserialize(buffer)
        return torch.from_numpy(np_array)

    recursive_serde_register(
        torch.Tensor,
        serialize=torch_serialize,
        deserialize=lambda data: torch_deserialize(data),
        canonical_name="torch_tensor",
        version=1,
    )

except ImportError:  # nosec
    pass

# unsure why we have to register the object not the type but this works
recursive_serde_register(
    np.core._ufunc_config._unspecified(),
    canonical_name="numpy_ufunc_unspecified",
    version=1,
)

recursive_serde_register(
    pydantic.EmailStr,
    serialize=lambda x: x.encode(),
    deserialize=lambda x: pydantic.EmailStr(x.decode()),
    canonical_name="pydantic_emailstr",
    version=1,
)


# how else do you import a relative file to execute it?
NOTHING = None

try:
    # Just register these serializers if the google.cloud.bigquery & db_dtypes module are available
    # third party
    from google.cloud.bigquery.job.query import QueryJob
    from google.cloud.bigquery.table import RowIterator

    # Checking db_dtypes availability this way to avoid unused ruff issues, but this package is used internally
    if not find_spec("db_dtypes"):
        raise ImportError("db_dtypes module not found")

    def convert_to_dataframe(obj: RowIterator) -> bytes:
        dataframe = obj.to_dataframe()
        return serialize_dataframe(dataframe)

    def convert_from_dataframe(blob: bytes) -> DataFrame:
        dataframe = deserialize_dataframe(blob)
        return dataframe

    recursive_serde_register(
        RowIterator,
        serialize=convert_to_dataframe,
        deserialize=convert_from_dataframe,
        canonical_name="bigquery_rowiterator",
        version=1,
    )

    recursive_serde_register(
        QueryJob,
        serialize=lambda obj: convert_to_dataframe(obj.result()),
        deserialize=convert_from_dataframe,
        canonical_name="bigquery_queryjob",
        version=1,
    )
except ImportError:
    pass



# stdlib
import ast
import inspect

# third party
from IPython import get_ipython

# relative
from ..response import SyftException
from ..response import SyftWarning
from .code_parse import GlobalsVisitor
from .code_parse import LaunchJobVisitor


def submit_subjobs_code(submit_user_code, ep_client) -> None:  # type: ignore
    # TODO: fix the mypy issue. Reason: circular import
    # We are exploring the source code to automatically upload
    # subjobs in the ephemeral server
    # Usually, a DS would manually submit the code for subjobs,
    # but because we dont allow them to interact with the ephemeral server
    # that would not be possible
    if "datasite" in submit_user_code.input_kwargs:
        tree = ast.parse(inspect.getsource(submit_user_code.local_function))
        v = LaunchJobVisitor()
        v.visit(tree)
        nested_calls = v.nested_calls
        try:
            ipython = (
                get_ipython()
            )  # works only in interactive envs (like jupyter notebooks)
        except Exception:
            ipython = None
            pass

        for call in nested_calls:
            if ipython is not None:
                specs = ipython.object_inspect(call)
                # Look for nested job locally, maybe we could
                # fetch
                if specs["type_name"] == "SubmitUserCode":
                    ep_client.code.submit(ipython.ev(call))


def check_for_global_vars(code_tree: ast.Module) -> GlobalsVisitor | SyftWarning:
    """
    Check that the code does not contain any global variables
    """
    v = GlobalsVisitor()
    try:
        v.visit(code_tree)
    except Exception:
        raise SyftException(
            "Your code contains (a) global variable(s), which is not allowed"
        )
    return v


def parse_code(raw_code: str) -> ast.Module | SyftWarning:
    """
    Parse the code into an AST tree and return a warning if there are syntax errors
    """
    try:
        tree = ast.parse(raw_code)
    except SyntaxError as e:
        raise SyftException(f"Your code contains syntax error: {e}")
    return tree



# stdlib
from collections.abc import Callable
from getpass import getpass
from typing import Any

# third party
from bcrypt import checkpw
from bcrypt import gensalt
from bcrypt import hashpw
from pydantic import EmailStr
from pydantic import ValidationError
from pydantic import field_validator

# relative
from ...client.api import APIRegistry
from ...serde.serializable import serializable
from ...server.credentials import SyftSigningKey
from ...server.credentials import SyftVerifyKey
from ...types.syft_metaclass import Empty
from ...types.syft_object import PartialSyftObject
from ...types.syft_object import SYFT_OBJECT_VERSION_1
from ...types.syft_object import SyftObject
from ...types.transforms import TransformContext
from ...types.transforms import drop
from ...types.transforms import generate_id
from ...types.transforms import keep
from ...types.transforms import make_set_default
from ...types.transforms import transform
from ...types.transforms import validate_email
from ...types.uid import UID
from ..notifier.notifier_enums import NOTIFIERS
from ..response import SyftError
from ..response import SyftSuccess
from .user_roles import ServiceRole


@serializable()
class User(SyftObject):
    # version
    __canonical_name__ = "User"
    __version__ = SYFT_OBJECT_VERSION_1

    id: UID | None = None  # type: ignore[assignment]

    # fields
    notifications_enabled: dict[NOTIFIERS, bool] = {
        NOTIFIERS.EMAIL: True,
        NOTIFIERS.SMS: False,
        NOTIFIERS.SLACK: False,
        NOTIFIERS.APP: False,
    }
    email: EmailStr | None = None
    name: str | None = None
    hashed_password: str | None = None
    salt: str | None = None
    signing_key: SyftSigningKey | None = None
    verify_key: SyftVerifyKey | None = None
    role: ServiceRole | None = None
    institution: str | None = None
    website: str | None = None
    created_at: str | None = None
    # TODO where do we put this flag?
    mock_execution_permission: bool = False

    # serde / storage rules
    __attr_searchable__ = ["name", "email", "verify_key", "role"]
    __attr_unique__ = ["email", "signing_key", "verify_key"]
    __repr_attrs__ = ["name", "email"]


def default_role(role: ServiceRole) -> Callable:
    return make_set_default(key="role", value=role)


def hash_password(context: TransformContext) -> TransformContext:
    if context.output is None:
        return context

    if context.output["password"] is not None and (
        (context.output["password_verify"] is None)
        or context.output["password"] == context.output["password_verify"]
    ):
        salt, hashed = salt_and_hash_password(context.output["password"], 12)
        context.output["hashed_password"] = hashed
        context.output["salt"] = salt

    return context


def generate_key(context: TransformContext) -> TransformContext:
    if context.output is not None:
        signing_key = SyftSigningKey.generate()
        context.output["signing_key"] = signing_key
        context.output["verify_key"] = signing_key.verify_key

    return context


def salt_and_hash_password(password: str, rounds: int) -> tuple[str, str]:
    bytes_pass = password.encode("UTF-8")
    salt = gensalt(rounds=rounds)
    hashed = hashpw(bytes_pass, salt)
    hashed_bytes = hashed.decode("UTF-8")
    salt_bytes = salt.decode("UTF-8")
    return salt_bytes, hashed_bytes


def check_pwd(password: str, hashed_password: str) -> bool:
    return checkpw(
        password=password.encode("utf-8"),
        hashed_password=hashed_password.encode("utf-8"),
    )


@serializable()
class UserUpdate(PartialSyftObject):
    __canonical_name__ = "UserUpdate"
    __version__ = SYFT_OBJECT_VERSION_1

    @field_validator("role", mode="before")
    @classmethod
    def str_to_role(cls, v: Any) -> Any:
        if isinstance(v, str) and hasattr(ServiceRole, v.upper()):
            return getattr(ServiceRole, v.upper())
        return v

    email: EmailStr
    name: str
    role: ServiceRole  # make sure role cant be set without uid
    password: str
    password_verify: str
    verify_key: SyftVerifyKey
    institution: str
    website: str
    mock_execution_permission: bool


@serializable()
class UserCreate(SyftObject):
    __canonical_name__ = "UserCreate"
    __version__ = SYFT_OBJECT_VERSION_1

    email: EmailStr
    name: str
    role: ServiceRole | None = None  # type: ignore[assignment]
    password: str
    password_verify: str | None = None  # type: ignore[assignment]
    verify_key: SyftVerifyKey | None = None  # type: ignore[assignment]
    institution: str | None = ""  # type: ignore[assignment]
    website: str | None = ""  # type: ignore[assignment]
    created_by: SyftSigningKey | None = None  # type: ignore[assignment]
    mock_execution_permission: bool = False

    __repr_attrs__ = ["name", "email"]


@serializable()
class UserSearch(PartialSyftObject):
    __canonical_name__ = "UserSearch"
    __version__ = SYFT_OBJECT_VERSION_1

    id: UID
    email: EmailStr
    verify_key: SyftVerifyKey
    name: str


@serializable()
class UserView(SyftObject):
    __canonical_name__ = "UserView"
    __version__ = SYFT_OBJECT_VERSION_1

    notifications_enabled: dict[NOTIFIERS, bool] = {
        NOTIFIERS.EMAIL: True,
        NOTIFIERS.SMS: False,
        NOTIFIERS.SLACK: False,
        NOTIFIERS.APP: False,
    }
    email: EmailStr
    name: str
    role: ServiceRole  # make sure role cant be set without uid
    institution: str | None = None
    website: str | None = None
    mock_execution_permission: bool

    __repr_attrs__ = [
        "name",
        "email",
        "institution",
        "website",
        "role",
        "notifications_enabled",
    ]

    def _coll_repr_(self) -> dict[str, Any]:
        return {
            "Name": self.name,
            "Email": self.email,
            "Institute": self.institution,
            "Website": self.website,
            "Role": self.role.name.capitalize(),
            "Notifications": "Email: "
            + (
                "Enabled" if self.notifications_enabled[NOTIFIERS.EMAIL] else "Disabled"
            ),
        }

    def _set_password(self, new_password: str) -> SyftError | SyftSuccess:
        api = APIRegistry.api_for(
            server_uid=self.syft_server_location,
            user_verify_key=self.syft_client_verify_key,
        )
        if api is None:
            return SyftError(message=f"You must login to {self.server_uid}")

        api.services.user.update(uid=self.id, password=new_password)
        return SyftSuccess(
            message=f"Successfully updated password for "
            f"user '{self.name}' with email '{self.email}'."
        )

    def set_password(
        self, new_password: str | None = None, confirm: bool = True
    ) -> SyftError | SyftSuccess:
        """Set a new password interactively with confirmed password from user input"""
        # TODO: Add password validation for special characters
        if not new_password:
            new_password = getpass("New Password: ")

        if confirm:
            confirmed_password: str = getpass("Please confirm your password: ")
            if confirmed_password != new_password:
                return SyftError(message="Passwords do not match !")
        return self._set_password(new_password)

    def set_email(self, email: str) -> SyftSuccess | SyftError:
        # validate email address
        api = APIRegistry.api_for(
            server_uid=self.syft_server_location,
            user_verify_key=self.syft_client_verify_key,
        )
        if api is None:
            return SyftError(message=f"You must login to {self.server_uid}")

        try:
            user_update = UserUpdate(email=email)
        except ValidationError:
            return SyftError(message="{email} is not a valid email address.")

        result = api.services.user.update(uid=self.id, email=user_update.email)

        if isinstance(result, SyftError):
            return result

        self.email = email
        return SyftSuccess(
            message=f"Successfully updated email for the user "
            f"'{self.name}' to '{self.email}'."
        )

    def update(
        self,
        name: type[Empty] | str = Empty,
        institution: type[Empty] | str = Empty,
        website: type[Empty] | str = Empty,
        role: type[Empty] | str = Empty,
        mock_execution_permission: type[Empty] | bool = Empty,
    ) -> SyftSuccess | SyftError:
        """Used to update name, institution, website of a user."""
        api = APIRegistry.api_for(
            server_uid=self.syft_server_location,
            user_verify_key=self.syft_client_verify_key,
        )
        if api is None:
            return SyftError(message=f"You must login to {self.server_uid}")
        user_update = UserUpdate(
            name=name,
            institution=institution,
            website=website,
            role=role,
            mock_execution_permission=mock_execution_permission,
        )
        result = api.services.user.update(uid=self.id, **user_update)

        if isinstance(result, SyftError):
            return result

        for attr, val in result.to_dict(exclude_empty=True).items():
            setattr(self, attr, val)

        return SyftSuccess(message="User details successfully updated.")

    def allow_mock_execution(self, allow: bool = True) -> SyftSuccess | SyftError:
        return self.update(mock_execution_permission=allow)


@serializable()
class UserViewPage(SyftObject):
    __canonical_name__ = "UserViewPage"
    __version__ = SYFT_OBJECT_VERSION_1

    users: list[UserView]
    total: int


@transform(UserUpdate, User)
def user_update_to_user() -> list[Callable]:
    return [
        validate_email,
        hash_password,
        drop(["password", "password_verify"]),
    ]


@transform(UserCreate, User)
def user_create_to_user() -> list[Callable]:
    return [
        generate_id,
        validate_email,
        hash_password,
        generate_key,
        drop(["password", "password_verify", "created_by"]),
        # TODO: Fix this by passing it from client & verifying it at server
        default_role(ServiceRole.DATA_SCIENTIST),
    ]


@transform(User, UserView)
def user_to_view_user() -> list[Callable]:
    return [
        keep(
            [
                "id",
                "email",
                "name",
                "role",
                "institution",
                "website",
                "mock_execution_permission",
                "notifications_enabled",
            ]
        )
    ]


@serializable()
class UserPrivateKey(SyftObject):
    __canonical_name__ = "UserPrivateKey"
    __version__ = SYFT_OBJECT_VERSION_1

    email: str
    signing_key: SyftSigningKey
    role: ServiceRole


@transform(User, UserPrivateKey)
def user_to_user_verify() -> list[Callable]:
    return [keep(["email", "signing_key", "id", "role"])]



# stdlib
from collections.abc import Callable
import logging
import multiprocessing
import multiprocessing.synchronize
import os
from pathlib import Path
import platform
import signal
import subprocess  # nosec
import sys
import time
from typing import Any

# third party
from fastapi import APIRouter
from fastapi import FastAPI
from pydantic_settings import BaseSettings
from pydantic_settings import SettingsConfigDict
import requests
from starlette.middleware.cors import CORSMiddleware
import uvicorn

# relative
from ..abstract_server import ServerSideType
from ..client.client import API_PATH
from ..util.autoreload import enable_autoreload
from ..util.constants import DEFAULT_TIMEOUT
from ..util.util import os_name
from .datasite import Datasite
from .enclave import Enclave
from .gateway import Gateway
from .routes import make_routes
from .server import ServerType
from .utils import get_named_server_uid
from .utils import remove_temp_dir_for_server

if os_name() == "macOS":
    # needed on MacOS to prevent [__NSCFConstantString initialize] may have been in
    # progress in another thread when fork() was called.
    multiprocessing.set_start_method("spawn", True)

WAIT_TIME_SECONDS = 20


class AppSettings(BaseSettings):
    name: str
    server_type: ServerType = ServerType.DATASITE
    server_side_type: ServerSideType = ServerSideType.HIGH_SIDE
    processes: int = 1
    reset: bool = False
    dev_mode: bool = False
    enable_warnings: bool = False
    in_memory_workers: bool = True
    queue_port: int | None = None
    create_producer: bool = False
    n_consumers: int = 0
    association_request_auto_approval: bool = False
    background_tasks: bool = False

    model_config = SettingsConfigDict(env_prefix="SYFT_", env_parse_none_str="None")


def app_factory() -> FastAPI:
    settings = AppSettings()

    worker_classes = {
        ServerType.DATASITE: Datasite,
        ServerType.GATEWAY: Gateway,
        ServerType.ENCLAVE: Enclave,
    }
    if settings.server_type not in worker_classes:
        raise NotImplementedError(
            f"server_type: {settings.server_type} is not supported"
        )
    worker_class = worker_classes[settings.server_type]

    kwargs = settings.model_dump()
    if settings.dev_mode:
        print(
            f"WARN: private key is based on server name: {settings.name} in dev_mode. "
            "Don't run this in production."
        )
        worker = worker_class.named(**kwargs)
    else:
        worker = worker_class(**kwargs)

    app = FastAPI(title=settings.name)
    router = make_routes(worker=worker)
    api_router = APIRouter()
    api_router.include_router(router)
    app.include_router(api_router, prefix="/api/v2")
    app.add_middleware(
        CORSMiddleware,
        allow_origins=["*"],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )
    return app


def attach_debugger() -> None:
    # third party
    import debugpy

    os.environ["PYDEVD_DISABLE_FILE_VALIDATION"] = "1"
    _, debug_port = debugpy.listen(0)
    print(
        "\nStarting the server with the Python Debugger enabled (`debug=True`).\n"
        'To attach the debugger, open the command palette in VSCode and select "Debug: Start Debugging (F5)".\n'
        f"Then, enter `{debug_port}` in the port field and press Enter.\n",
        flush=True,
    )
    print(f"Waiting for debugger to attach on port `{debug_port}`...", flush=True)
    debugpy.wait_for_client()  # blocks execution until a remote debugger is attached
    print("Debugger attached", flush=True)


def run_uvicorn(
    host: str,
    port: int,
    starting_uvicorn_event: multiprocessing.synchronize.Event,
    **kwargs: Any,
) -> None:
    dev_mode = kwargs.get("dev_mode")
    should_reset = dev_mode and kwargs.get("reset")

    if should_reset:
        print("Found `reset=True` in the launch configuration. Resetting the server...")
        named_server_uid = get_named_server_uid(kwargs.get("name"))
        remove_temp_dir_for_server(named_server_uid)
        # Explicitly set `reset` to False to prevent multiple resets during hot-reload
        kwargs["reset"] = False
        # Kill all old python processes
        try:
            python_pids = find_python_processes_on_port(port)
            for pid in python_pids:
                print(f"Stopping process on port: {port}")
                kill_process(pid)
                time.sleep(1)
        except Exception:  # nosec
            print(f"Failed to kill python process on port: {port}")

    log_level = "critical"
    if dev_mode:
        log_level = "info"
        logging.getLogger("uvicorn").setLevel(logging.CRITICAL)
        logging.getLogger("uvicorn.access").setLevel(logging.CRITICAL)

    if kwargs.get("debug"):
        attach_debugger()

    # Set up all kwargs as environment variables so that they can be accessed in the app_factory function.
    env_prefix = AppSettings.model_config.get("env_prefix", "")
    for key, value in kwargs.items():
        key_with_prefix = f"{env_prefix}{key.upper()}"
        os.environ[key_with_prefix] = str(value)

    # The `serve_server` function calls `run_uvicorn` in a separate process using `multiprocessing.Process`.
    # When the child process is created, it inherits the file descriptors from the parent process.
    # If the parent process has a file descriptor open for sys.stdin, the child process will also have a file descriptor
    # open for sys.stdin. This can cause an OSError in uvicorn when it tries to access sys.stdin in the child process.
    # To prevent this, we set sys.stdin to None in the child process. This is safe because we don't actually need
    # sys.stdin while running uvicorn programmatically.
    sys.stdin = None  # type: ignore

    # Signal the parent process that we are starting the uvicorn server.
    starting_uvicorn_event.set()

    # Finally, run the uvicorn server.
    uvicorn.run(
        "syft.server.uvicorn:app_factory",
        host=host,
        port=port,
        factory=True,
        reload=dev_mode,
        reload_dirs=[Path(__file__).parent.parent] if dev_mode else None,
        log_level=log_level,
    )


def serve_server(
    name: str,
    server_type: ServerType = ServerType.DATASITE,
    server_side_type: ServerSideType = ServerSideType.HIGH_SIDE,
    host: str = "0.0.0.0",  # nosec
    port: int = 8080,
    processes: int = 1,
    reset: bool = False,
    dev_mode: bool = False,
    tail: bool = False,
    enable_warnings: bool = False,
    in_memory_workers: bool = True,
    queue_port: int | None = None,
    create_producer: bool = False,
    n_consumers: int = 0,
    association_request_auto_approval: bool = False,
    background_tasks: bool = False,
    debug: bool = False,
) -> tuple[Callable, Callable]:
    starting_uvicorn_event = multiprocessing.Event()

    # Enable IPython autoreload if dev_mode is enabled.
    if dev_mode:
        enable_autoreload()

    server_process = multiprocessing.Process(
        target=run_uvicorn,
        kwargs={
            "name": name,
            "server_type": server_type,
            "host": host,
            "port": port,
            "processes": processes,
            "reset": reset,
            "dev_mode": dev_mode,
            "server_side_type": server_side_type,
            "enable_warnings": enable_warnings,
            "in_memory_workers": in_memory_workers,
            "queue_port": queue_port,
            "create_producer": create_producer,
            "n_consumers": n_consumers,
            "association_request_auto_approval": association_request_auto_approval,
            "background_tasks": background_tasks,
            "debug": debug,
            "starting_uvicorn_event": starting_uvicorn_event,
        },
    )

    def stop() -> None:
        print(f"Stopping {name}")
        server_process.terminate()
        server_process.join(3)
        if server_process.is_alive():
            # this is needed because often the process is still alive
            server_process.kill()
            print("killed")

    def start() -> None:
        print(f"Starting {name} server on {host}:{port}")
        server_process.start()

        # Wait for the child process to start uvicorn server before starting the readiness checks.
        starting_uvicorn_event.wait()

        if tail:
            try:
                while True:
                    time.sleep(1)
            except KeyboardInterrupt:
                try:
                    stop()
                except SystemExit:
                    os._exit(130)
        else:
            for i in range(WAIT_TIME_SECONDS):
                try:
                    req = requests.get(
                        f"http://{host}:{port}{API_PATH}/metadata",
                        timeout=DEFAULT_TIMEOUT,
                    )
                    if req.status_code == 200:
                        print(" Done.")
                        break
                except Exception:
                    time.sleep(1)
                    if i == 0:
                        print("Waiting for server to start", end="")
                    else:
                        print(".", end="")

    return start, stop


def find_python_processes_on_port(port: int) -> list[int]:
    system = platform.system()

    if system == "Windows":
        command = f"netstat -ano | findstr :{port}"
        process = subprocess.Popen(  # nosec
            command,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        output, _ = process.communicate()
        pids = [
            int(line.strip().split()[-1]) for line in output.split("\n") if line.strip()
        ]

    else:  # Linux and MacOS
        command = f"lsof -i :{port} -sTCP:LISTEN -t"
        process = subprocess.Popen(  # nosec
            command,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        output, _ = process.communicate()
        pids = [int(pid.strip()) for pid in output.split("\n") if pid.strip()]

    python_pids = []
    for pid in pids:
        if system == "Windows":
            command = (
                f"wmic process where (ProcessId='{pid}') get ProcessId,CommandLine"
            )
        else:
            command = f"ps -p {pid} -o pid,command"

        try:
            process = subprocess.Popen(  # nosec
                command,
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
            )
            output, _ = process.communicate()
        except Exception as e:
            print(f"Error checking process {pid}: {e}")
            continue

        lines = output.strip().split("\n")
        if len(lines) > 1 and "python" in lines[1].lower():
            python_pids.append(pid)

    return python_pids


def kill_process(pid: int) -> None:
    try:
        os.kill(pid, signal.SIGTERM)
        print(f"Process {pid} terminated.")
    except Exception as e:
        print(f"Error killing process {pid}: {e}")
